# STEclat is one of the fundamental algorithm to discover geo refereneced partial periodic-frequent patterns in a transactional database.
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#             import PAMI.georeferencedPartialPeriodicPattern.STEclat as alg
#
#             obj = alg.STEclat("sampleTDB.txt", "sampleN.txt", 3, 4)
#
#             obj.mine()
#
#             partialPeriodicSpatialPatterns = obj.getPatterns()
#
#             print("Total number of Periodic Spatial Frequent Patterns:", len(partialPeriodicSpatialPatterns))
#
#             obj.save("outFile")
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#




__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""


from PAMI.georeferencedPartialPeriodicPattern.basic import abstract as _ab
from deprecated import deprecated


class STEclat(_ab._partialPeriodicSpatialPatterns):
    """
    :Description:   STEclat is one of the fundamental algorithm to discover georefereneced partial periodic-frequent patterns in a transactional database.

    :Reference:   R. Uday Kiran, C. Saideep, K. Zettsu, M. Toyoda, M. Kitsuregawa and P. Krishna Reddy,
                 "Discovering Partial Periodic Spatial Patterns in Spatiotemporal Databases," 2019 IEEE International
                  Conference on Big Data (Big Data), 2019, pp. 233-238, doi: 10.1109/BigData47090.2019.9005693.

    :param  iFile: str :
                   Name of the Input file to mine complete set of Geo-referenced Partial Periodic patterns
    :param  oFile: str :
                   Name of the output file to store complete set of Geo-referenced Partial Periodic patterns
    :param  minPS: int or float or str :
                   The user can specify minPS either in count or proportion of database size. If the program detects the data type of minPS is integer, then it treats minPS is expressed in count. Otherwise, it will be treated as float.
    :param maxIAT: int or float or str :
                   The user can specify maxIAT either in count or proportion of database size. If the program detects the data type of maxIAT is integer, then it treats maxIAT is expressed in count. Otherwise, it will be treated as float.
    :param nFile: str :
                   Name of the input file to mine complete set of Geo-referenced Partial Periodic patterns
    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.

    :Attributes:

        iFile : str
            Input file name or path of the input file
        nFile : str
           Name of Neighbourhood file name
        maxIAT : float or int or str
            The user can specify maxIAT either in count or proportion of database size.
            If the program detects the data type of maxIAT is integer, then it treats maxIAT is expressed in count.
            Otherwise, it will be treated as float.
            Example: maxIAT=10 will be treated as integer, while maxIAT=10.0 will be treated as float
        minPS : float or int or str
            The user can specify minPS either in count or proportion of database size.
            If the program detects the data type of minPS is integer, then it treats minPS is expressed in count.
            Otherwise, it will be treated as float.
            Example: minPS=10 will be treated as integer, while minPS=10.0 will be treated as float
        sep : str
            This variable is used to distinguish items from one another in a transaction. The default separator is tab space or \t.
            However, the users can override their default separator.
        startTime : float
            To record the start time of the mining process
        endTime : float
            To record the completion time of the mining process
        finalPatterns : dict
            Storing the complete set of patterns in a dictionary variable
        oFile : str
            Name of the output file to store complete set of frequent patterns
        memoryUSS : float
            To store the total amount of USS memory consumed by the program
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program
        Database : list
            To store the complete set of transactions available in the input database/file

    :Methods:

        mine()
            Mining process will start from here
        getPatterns()
            Complete set of patterns will be retrieved with this function
        save(oFile)
            Complete set of frequent patterns will be loaded in to a output file
        getPatternsAsDataFrames()
            Complete set of frequent patterns will be loaded in to a dataframe
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function
        creatingItemSets(iFileName)
            Storing the complete transactions of the database/input file in a database variable
        frequentOneItem()
            Generating one frequent patterns
        convert(value):
            To convert the given user specified value
        getNeighbourItems(keySet)
            A function to get common neighbours of a itemSet
        mapNeighbours(file)
            A function to map items to their neighbours

    **Executing the code on terminal :**
    ----------------------------------------

    .. code-block:: console

      Format:

      (.venv) $ python3 STEclat.py <inputFile> <outputFile> <neighbourFile>  <minPS>  <maxIAT>

      Example Usage:

      (.venv) $ python3 STEclat.py sampleTDB.txt output.txt sampleN.txt 0.2 0.5

    .. note:: maxIAT & minPS will be considered in percentage of database transactions


    **Sample run of importing the code :**
    --------------------------------------
    .. code-block:: python
    
            import PAMI.georeferencedPartialPeriodicPattern.STEclat as alg

            obj = alg.STEclat("sampleTDB.txt", "sampleN.txt", 3, 4)

            obj.mine()

            partialPeriodicSpatialPatterns = obj.getPatterns()

            print("Total number of Periodic Spatial Frequent Patterns:", len(partialPeriodicSpatialPatterns))

            obj.save("outFile")

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

    **Credits:**
    -------------
        The complete program was written by P. Likhitha under the supervision of Professor Rage Uday Kiran.
    """

    _maxIAT = " "
    _minPS = " "
    _startTime = float()
    _endTime = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _nFile = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []
    _sep = "\t"
    _lno = 0

    def __init__(self, iFile, nFile, minPS, maxIAT, sep="\t"):
        super().__init__(iFile, nFile, minPS, maxIAT,  sep)
        self._NeighboursMap = {}

    def _creatingItemSets(self):
        """
        Storing the complete transactions of the database/input file in a database variable
        """
        self._Database = []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            data, ts = [], []
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'TS' in i:
                ts = self._iFile['TS'].tolist()
            if 'Transactions' in i:
                data = self._iFile['Transactions'].tolist()
            for i in range(len(data)):
                tr = [ts[i][0]]
                tr = tr + data[i]
                self._Database.append(tr)

        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self._Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()

    # function to get frequent one pattern
    def _frequentOneItem(self):
        """
        Generating one frequent patterns
        """
        self._tidList = {}
        self._mapSupport = {}
        self._maxIAT = self._convert(self._maxIAT)
        for line in self._Database:
            s = line
            n = int(s[0])
            for i in range(1, len(s)):
                si = s[i]
                if self._mapSupport.get(si) is None:
                    self._mapSupport[si] = [0, n]
                    self._tidList[si] = [n]
                else:
                    lp = n - self._mapSupport[si][1]
                    if lp <= self._maxIAT:
                        self._mapSupport[si][0] += 1
                    self._mapSupport[si][1] = n
                    self._tidList[si].append(n)
        self._minPS = self._convert(self._minPS)
        self._mapSupport = {k: v[0] for k, v in self._mapSupport.items() if v[0] >= self._minPS}
        plist = [key for key, value in sorted(self._mapSupport.items(), key=lambda x: x[1], reverse=True)]
        return plist

    def _convert(self, value):
        """
        To convert the given user specified value

        :param value: user specified value

        :type value: int or float or str

        :return: converted value

        :rtype: float
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self._Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self._Database) * value)
            else:
                value = int(value)
        return value

    def _getPeriodicSupport(self, timeStamps):
        """
        calculates the support and periodicity with list of timestamps

        :param timeStamps: timestamps of a pattern
        :type timeStamps: list
        """
        timeStamps.sort()
        per = 0
        for i in range(len(timeStamps) - 1):
            j = i + 1
            if abs(timeStamps[j] - timeStamps[i]) <= self._maxIAT:
                per += 1
        return per

    def _save(self, prefix, suffix, tidSetX):
        """
        Saves the patterns that satisfy the periodic frequent property.

        :param prefix: the prefix of a pattern
        :type prefix: list or None
        :param suffix: the suffix of a patterns
        :type suffix: list
        :param tidSetX: the timestamp of a patterns
        :type tidSetX: list

        """
        if prefix is None:
            prefix = suffix
        else:
            prefix = prefix + suffix
        val = self._getPeriodicSupport(tidSetX)
        if val >= self._minPS:
            self._finalPatterns[tuple(prefix)] = val

    def _Generation(self, prefix, itemSets, tidSets):
        """
        Generates the patterns that satisfy the periodic frequent property.

        :param prefix: the prefix of a pattern
        :type prefix: list or None
        :param itemSets: the item sets of a patterns
        :type itemSets: list
        :param tidSets: the timestamp of a patterns
        :type tidSets: list
        """
        if len(itemSets) == 1:
            i = itemSets[0]
            tidi = tidSets[0]
            self._save(prefix, [i], tidi)
            return
        for i in range(len(itemSets)):
            itemI = itemSets[i]
            if itemI is None:
                continue
            tidSetX = tidSets[i]
            classItemSets = []
            classTidSets = []
            itemSetX = [itemI]
            for j in range(i + 1, len(itemSets)):
                itemJ = itemSets[j]
                tidSetJ = tidSets[j]
                y = list(set(tidSetX).intersection(tidSetJ))
                val = self._getPeriodicSupport(y)
                if val >= self._minPS:
                    classItemSets.append(itemJ)
                    classTidSets.append(y)
            newprefix = list(set(itemSetX)) + prefix
            self._Generation(newprefix, classItemSets, classTidSets)
            self._save(prefix, list(set(itemSetX)), tidSetX)

    def _getNeighbourItems(self, keySet):
        """
        A function to get Neighbours of an item

        :param keySet: itemSet
        :type keySet: str or tuple
        :return: set of common neighbours
        :rtype: set
        """
        itemNeighbours = self._NeighboursMap.keys()
        if isinstance(keySet, str):
            if self._NeighboursMap.get(keySet) is None:
                return []
            itemNeighbours = list(set(itemNeighbours).intersection(set(self._NeighboursMap.get(keySet))))
        if isinstance(keySet, tuple):
            keySet = list(keySet)
            for j in range(0, len(keySet)):
                i = keySet[j]
                itemNeighbours = list(set(itemNeighbours).intersection(set(self._NeighboursMap.get(i))))
        return itemNeighbours

    def mapNeighbours(self):
        """
        A function to map items to their Neighbours
        """
        self._NeighboursMap = {}
        if isinstance(self._nFile, _ab._pd.DataFrame):
            data = []
            if self._nFile.empty:
                print("its empty..")
            i = self._nFile.columns.values.tolist()
            if 'Neighbours' in i:
                data = self._nFile['Neighbours'].tolist()
            for i in data:
                self._NeighboursMap[i[0]] = i[1:]
        if isinstance(self._nFile, str):
            if _ab._validators.url(self._nFile):
                data = _ab._urlopen(self._nFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._NeighboursMap[temp[0]] = temp[1:]
            else:
                try:
                    with open(self._nFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self._NeighboursMap[temp[0]] = temp[1:]
                except IOError:
                    print("File Not Found")
                    quit()

    @deprecated("It is recommended to use 'mine()' instead of 'mine()' for mining process. Starting from January 2025, 'mine()' will be completely terminated.")
    def startMine(self):
        """
        Frequent pattern mining process will start from here
        """

        self.mine()

    def mine(self):
        """
        Frequent pattern mining process will start from here
        """

        # global items_sets, endTime, startTime
        self._startTime = _ab._time.time()
        if self._iFile is None:
            raise Exception("Please enter the file path or file name:")
        self._creatingItemSets()
        #self._minSup = self._convert(self._minSup)
        self.mapNeighbours()
        self._finalPatterns = {}
        plist = self._frequentOneItem()
        for i in range(len(plist)):
            itemX = plist[i]
            tidSetX = self._tidList[itemX]
            itemSetX = [itemX]
            itemSets = []
            tidSets = []
            neighboursItems = self._getNeighbourItems(plist[i])
            for j in range(i + 1, len(plist)):
                if not plist[j] in neighboursItems:
                    continue
                itemJ = plist[j]
                tidSetJ = self._tidList[itemJ]
                y1 = list(set(tidSetX).intersection(tidSetJ))
                val = self._getPeriodicSupport(y1)
                if val >= self._minPS:
                    itemSets.append(itemJ)
                    tidSets.append(y1)
            self._Generation(itemSetX, itemSets, tidSets)
            self._save(None, itemSetX, tidSetX)
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Spatial Periodic Frequent patterns were generated successfully using SpatialEclat algorithm")

    def getMemoryUSS(self):
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """
        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """
        Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            pat = ""
            for i in a:
                pat += str(i) + ' '
            data.append([pat, b])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'periodicSupport'])
        return dataFrame

    def save(self, outFile):
        """
        Complete set of frequent patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: csv file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            pat = ""
            for i in x:
                pat += str(i) + '\t'
            patternsAndSupport = pat.strip() + ": " + str(y)
            writer.write("%s \n" % patternsAndSupport)

    def getPatterns(self):
        """
        Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self):
        """
        This function is used to print the results
        """
        print("Total number of  Spatial Partial Periodic Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:",  self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 6 or len(_ab._sys.argv) == 7:
        if len(_ab._sys.argv) == 7:
            _ap = STEclat(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5], _ab._sys.argv[6])
        if len(_ab._sys.argv) == 6:
            _ap = STEclat(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5])
        _ap.mine()
        _ap.mine()
        print("Total number of Spatial Frequent Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in seconds:",  _ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")






# Local Periodic Patterns, which are patterns (sets of events) that have a periodic behavior in some non predefined
# time-intervals. A pattern is said to be a local periodic pattern if it appears regularly and continuously in some
# time-intervals. The maxSoPer (maximal period of spillovers) measure allows detecting time-intervals of variable
# lengths where a pattern is continuously periodic, while the minDur (minimal duration) measure ensures that those
# time-intervals have a minimum duration.
#
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#             from PAMI.localPeriodicPattern.basic import LPPGrowth as alg
#
#             obj = alg.LPPGrowth(iFile, maxPer, maxSoPer, minDur)
#
#             obj.mine()
#
#             localPeriodicPatterns = obj.getPatterns()
#
#             print(f'Total number of local periodic patterns: {len(localPeriodicPatterns)}')
#
#             obj.save(oFile)
#
#             Df = obj.getPatternsAsDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print(f'Total memory in USS: {memUSS}')
#
#             memRSS = obj.getMemoryRSS()
#
#             print(f'Total memory in RSS: {memRSS}')
#
#             runtime = obj.getRuntime()
#
#             print(f'Total execution time in seconds: {runtime})
#





__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""



from PAMI.localPeriodicPattern.basic import abstract as _ab
from typing import List, Dict, Tuple, Set, Union, Any, Generator
from deprecated import deprecated

class Node:
    """
    A class used to represent the node of localPeriodicPatternTree

    :Attributes:

        item : int
            storing item of a node
        parent : node
            To maintain the parent of every node
        child : list
            To maintain the children of node
        nodeLink : node
            To maintain the next node of node
        tidList : set
            To maintain timestamps of node

    :Methods:

        getChild(itemName)
            storing the children to their respective parent nodes
    """
    
    def __init__(self) -> None:
        self.item = -1
        self.parent = None
        self.child = []
        self.nodeLink = None
        self.tidList = set()

    def getChild(self, item: int) -> 'Node':
        """
        This function is used to get child node from the parent node

        :param item: item of the parent node

        :type item: int

        :return: if node have node of item, then return it. if node don't have return []

        :rtype: Node
        """
        for child in self.child:
            if child.item == item:
                return child
        return []


class Tree:
    """
        A class used to represent the frequentPatternGrowth tree structure

        :Attributes:

            root : node
                Represents the root node of the tree
            nodeLinks : dictionary
                storing last node of each item
            firstNodeLink : dictionary
                storing first node of each item

        :Methods:

            addTransaction(transaction,timeStamp)
                creating transaction as a branch in frequentPatternTree
            fixNodeLinks(itemName, newNode)
                add newNode link after last node of item
            deleteNode(itemName)
                delete all node of item
            createPrefixTree(path,timeStampList)
                create prefix tree by path

    """
    def __init__(self) -> None:
        self.root = Node()
        self.nodeLinks = {}
        self.firstNodeLink = {}

    def addTransaction(self, transaction: List[int], tid: int) -> None:
        """
        add transaction into tree

        :param transaction: it represents the one transaction in database
        :type transaction: list
        :param tid: represents the timestamp of transaction
        :type tid: list or int
        :return: None
        """
        current = self.root
        for item in transaction:
            child = current.getChild(item)
            if not child:
                newNode = Node()
                newNode.item = item
                newNode.parent = current
                current.child.append(newNode)
                current = newNode
                self.fixNodeLinks(item, newNode)
            else:
                current = child
            current.tidList.add(tid)

    def fixNodeLinks(self, item: int, newNode: 'Node') -> None:
        """
        fix node link

        :param item: it represents item name of newNode
        :type item: string
        :param newNode: it represents node which is added
        :type newNode: Node
        :return: None
        """
        if item in self.nodeLinks:
            lastNode = self.nodeLinks[item]
            lastNode.nodeLink = newNode
        self.nodeLinks[item] = newNode
        if item not in self.firstNodeLink:
            self.firstNodeLink[item] = newNode

    def deleteNode(self, item: int) -> None:
        """
        delete the node from tree

        :param item: it represents the item name of node
        :type item: str
        :return: None
        """
        deleteNode = self.firstNodeLink[item]
        parentNode = deleteNode.parent
        parentNode.child.remove(deleteNode)
        parentNode.child += deleteNode.child
        parentNode.tidList |= deleteNode.tidList
        for child in deleteNode.child:
            child.parent = parentNode
        while deleteNode.nodeLink:
            deleteNode = deleteNode.nodeLink
            parentNode = deleteNode.parent
            parentNode.child.remove(deleteNode)
            parentNode.child += deleteNode.child
            parentNode.tidList |= deleteNode.tidList
            for child in deleteNode.child:
                child.parent = parentNode

    def createPrefixTree(self, path: List[int], tidList: List[int]) -> None:
        """
        create prefix tree by path

        :param path: it represents path to root from prefix node
        :type path: list
        :param tidList: it represents tid of each item
        :type tidList: list
        :return: None
        """
        currentNode = self.root
        for item in path:
            child = currentNode.getChild(item)
            if not child:
                newNode = Node()
                newNode.item = item
                newNode.parent = currentNode
                currentNode.child.append(newNode)
                currentNode = newNode
                self.fixNodeLinks(item, newNode)
            else:
                currentNode = child
            currentNode.tidList |= tidList


class LPPGrowth(_ab._localPeriodicPatterns):
    """
    :Description:

        Local Periodic Patterns, which are patterns (sets of events) that have a periodic behavior in some non predefined
        time-intervals. A pattern is said to be a local periodic pattern if it appears regularly and continuously in some
        time-intervals. The maxSoPer (maximal period of spillovers) measure allows detecting time-intervals of variable
        lengths where a pattern is continuously periodic, while the minDur (minimal duration) measure ensures that those
        time-intervals have a minimum duration.

    :Reference:

        Fournier-Viger, P., Yang, P., Kiran, R. U., Ventura, S., Luna, J. M.. (2020). Mining Local Periodic Patterns in
        a Discrete Sequence. Information Sciences, Elsevier, to appear. [ppt] DOI: 10.1016/j.ins.2020.09.044

    :param  iFile: str :
                   Name of the Input file to mine complete set of local periodic pattern's
    :param  oFile: str :
                   Name of the output file to store complete set of local periodic patterns
    :param  minDur: str:
                   Minimal duration in seconds between consecutive periods of time-intervals where a pattern is continuously periodic.
    :param  maxPer: float:
                   Controls the maximum number of transactions in which any two items within a pattern can reappear.
    :param  maxSoPer: float:
                   Controls the maximum number of time periods between consecutive periods of time-intervals where a pattern is continuously periodic.

    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.

    :Attributes:

            iFile : str
                Input file name or path of the input file
            oFile : str
                Output file name or path of the output file
            maxPer : float
                User defined maxPer value.
            maxSoPer : float
                User defined maxSoPer value.
            minDur : float
                User defined minDur value.
            tsMin : int / date
                First time stamp of input data.
            tsMax : int / date
                Last time stamp of input data.
            startTime : float
                Time when start of execution the algorithm.
            endTime : float
                Time when end of execution the algorithm.
            finalPatterns : dict
                To store local periodic patterns and its PTL.
            tsList : dict
                To store items and its time stamp as bit vector.
            root : Tree
                It is root node of transaction tree of whole input data.
            PTL : dict
                Storing the item and its PTL.
            items : list
                Storing local periodic item list.
            sep: str
                separator used to distinguish items from each other. The default separator is tab space.

    :Methods:

            findSeparator(line)
                Find the separator of the line which split strings.
            creteLPPlist()
                Create the local periodic patterns list from input data.
            createTSList()
                Create the tsList as bit vector from input data.
            generateLPP()
                Generate 1 length local periodic pattens by tsList and execute depth first search.
            createLPPTree()
                Create LPPTree of local periodic item from input data.
            patternGrowth(tree, prefix, prefixPFList)
                Execute pattern growth algorithm. It is important function in this program.
            calculatePTL(tsList)
                Calculate PTL from input tsList as integer list.
            calculatePTLbit(tsList)
                Calculate PTL from input tsList as bit vector.
            mine()
                Mining process will start from here.
            getMemoryUSS()
                Total amount of USS memory consumed by the mining process will be retrieved from this function.
            getMemoryRSS()
                Total amount of RSS memory consumed by the mining process will be retrieved from this function.
            getRuntime()
                Total amount of runtime taken by the mining process will be retrieved from this function.
            getLocalPeriodicPatterns()
                return local periodic patterns and its PTL
            save(oFile)
                Complete set of local periodic patterns will be loaded in to an output file.
            getPatternsAsDataFrame()
                Complete set of local periodic patterns will be loaded in to a dataframe.

    **Executing the code on terminal:**
    ---------------------------------------

    .. code-block:: console

      Format:

      (.venv) $ python3 LPPMGrowth.py <inputFile> <outputFile> <maxPer> <minSoPer> <minDur>

      Example Usage:

      (.venv) $ python3 LPPMGrowth.py sampleDB.txt patterns.txt 0.3 0.4 0.5

    .. note: minDur will be considered as time interval between two consecutive periods


    **Sample run of importing the code:**
    ----------------------------------------
    .. code-block:: python

            from PAMI.localPeriodicPattern.basic import LPPGrowth as alg

            obj = alg.LPPGrowth(iFile, maxPer, maxSoPer, minDur)

            obj.mine()

            localPeriodicPatterns = obj.getPatterns()

            print(f'Total number of local periodic patterns: {len(localPeriodicPatterns)}')

            obj.save(oFile)

            Df = obj.getPatternsAsDataFrame()

            memUSS = obj.getMemoryUSS()

            print(f'Total memory in USS: {memUSS}')

            memRSS = obj.getMemoryRSS()

            print(f'Total memory in RSS: {memRSS}')

            runtime = obj.getRuntime()

            print(f'Total execution time in seconds: {runtime})

    **Credits:**
    --------------
        The complete program was written by So Nakamura under the supervision of Professor Rage Uday Kiran.
    """
    _localPeriodicPatterns__iFile = ' '
    _localPeriodicPatterns__oFile = ' '
    _localPeriodicPatterns__maxPer = str()
    _localPeriodicPatterns__maxSoPer = str()
    _localPeriodicPatterns__minDur = str()
    __tsMin = 0
    __tsMax = 0
    _localPeriodicPatterns__startTime = float()
    _localPeriodicPatterns__endTime = float()
    _localPeriodicPatterns__memoryUSS = float()
    _localPeriodicPatterns__memoryRSS = float()
    _localPeriodicPatterns__finalPatterns = {}
    __tsList = {}
    __root = Tree()
    __PTL = {}
    __items = []
    _localPeriodicPatterns__sep = ' '
    __Database = []

    def __creatingItemSets(self) -> None:
        """
        Storing the complete transactions of the database/input file in a database variable
        """
        self.__Database = []
        if isinstance(self._localPeriodicPatterns__iFile, _ab._pd.DataFrame):
            if self._localPeriodicPatterns__iFile.empty:
                print("its empty..")
            i = self._localPeriodicPatterns__iFile.columns.values.tolist()
            if 'Transactions' in i:
                self.__Database = self._localPeriodicPatterns__iFile['Transactions'].tolist()
            if 'Patterns' in i:
                self.__Database = self._localPeriodicPatterns__iFile['Patterns'].tolist()

        if isinstance(self._localPeriodicPatterns__iFile, str):
            if _ab._validators.url(self._localPeriodicPatterns__iFile):
                data = _ab._urlopen(self._localPeriodicPatterns__iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._localPeriodicPatterns__sep)]
                    temp = [x for x in temp if x]
                    self.__Database.append(temp)
            else:
                try:
                    with open(self._localPeriodicPatterns__iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._localPeriodicPatterns__sep)]
                            temp = [x for x in temp if x]
                            self.__Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()

    def __createLPPlist(self) -> None:
        """
        Create Local Periodic Pattern list from temporal data.
        """
        LPPList = {}
        PTL = {}
        start = {}
        tsPre = {}
        for line in self.__Database:
            soPer = ' '
            self.__tsMin = int(line.pop(0))
            ts = self.__tsMin
            for item in line:
                if item in LPPList:
                    per = ts - tsPre[item]
                    if per <= self._localPeriodicPatterns__maxPer and start == -1:
                        start = tsPre[item]
                        soPer = self._localPeriodicPatterns__maxSoPer
                    if start != -1:
                        soPer = max(0, soPer + per - self._localPeriodicPatterns__maxPer)
                        if soPer > self._localPeriodicPatterns__maxSoPer:
                            if tsPre[item] - start[item] <= self._localPeriodicPatterns__minDur:
                                PTL[item].add((start[item], tsPre[item]))
                                LPPList[item] = PTL[item]
                            start[item] = -1
                else:
                    tsPre[item] = ts
                    start[item] = -1
                    LPPList[item] = set()
            for line in self.__Database:
                ts = int(line.pop(0))
                for item in line:
                    if item in LPPList:
                        per = ts - tsPre[item]
                        if per <= self._localPeriodicPatterns__maxPer and start[item] == -1:
                            start[item] = tsPre[item]
                            soPer = self._localPeriodicPatterns__maxSoPer
                        if start[item] != -1:
                            soPer = max(0, soPer + per - self._localPeriodicPatterns__maxPer)
                            if soPer > self._localPeriodicPatterns__maxSoPer:
                                PTL[item].add((start[item], tsPre[item]))
                                LPPList[item] = PTL[item]
                            start[item] = -1
                        tsPre[item] = ts
                    else:
                        tsPre[item] = ts
                        start[item] = -1
                        LPPList[item] = set()

    def __createTSList(self) -> None:
        """
        Create tsList as bit vector from temporal data.
        """
        # for line in self.__Database:
        #     count = 1
        #     bitVector = 0b1 << count
        #     bitVector = bitVector | 0b1
        #     self.__tsMin = int(line.pop(0))
        #     self.__tsList = {item: bitVector for item in line}
        #     count += 1
        #     ts = ' '
        count = 1
        for line in self.__Database:
            bitVector = 0b1 << count
            bitVector = bitVector | 0b1
            # print(line)
            ts = line[0]
            for item in line[1:]:
                if self.__tsList.get(item):
                    different = abs(bitVector.bit_length() - self.__tsList[item].bit_length())
                    self.__tsList[item] = self.__tsList[item] << different
                    self.__tsList[item] = self.__tsList[item] | 0b1
                else:
                    self.__tsList[item] = bitVector
            count += 1
            self.__tsMax = int(ts)

        for item in self.__tsList:
            different = abs(bitVector.bit_length() - self.__tsList[item].bit_length())
            self.__tsList[item] = self.__tsList[item] << different
        self._localPeriodicPatterns__maxPer = self.__convert(self._localPeriodicPatterns__maxPer)
        self._localPeriodicPatterns__maxSoPer = self.__convert(self._localPeriodicPatterns__maxSoPer)
        self._localPeriodicPatterns__minDur = self.__convert(self._localPeriodicPatterns__minDur)

    def __generateLPP(self) -> None:
        """
        Generate local periodic items from bit vector tsList.
        """
        PTL = {}
        for item in self.__tsList:
            PTL[item] = set()
            ts = list(bin(self.__tsList[item]))
            ts = ts[2:]
            start = -1
            currentTs = 1
            soPer = ' '
            tsPre = ' '
            for t in ts[currentTs:]:
                if t == '0':
                    currentTs += 1
                    continue
                else:
                    tsPre = currentTs
                    currentTs += 1
                    break
            for t in ts[currentTs:]:
                if t == '0':
                    currentTs += 1
                    continue
                else:
                    per = currentTs - tsPre
                    if per <= self._localPeriodicPatterns__maxPer and start == -1:
                        start = tsPre
                        soPer = self._localPeriodicPatterns__maxSoPer
                    if start != -1:
                        soPer = max(0, soPer + per - self._localPeriodicPatterns__maxPer)
                        if soPer > self._localPeriodicPatterns__maxSoPer:
                            if tsPre - start >= self._localPeriodicPatterns__minDur:
                                PTL[item].add((start, tsPre))
                            """else:
                                bitVector = 0b1 << currentTs
                                different = abs(self.tsList[item].bit_length() - bitVector.bit_length())
                                bitVector = bitVector | 0b1
                                bitVector = bitVector << different
                                self.tsList[item] = self.tsList[item] | bitVector"""
                            start = -1
                    tsPre = currentTs
                    currentTs += 1
            if start != -1:
                soPer = max(0, soPer + self.__tsMax - tsPre - self._localPeriodicPatterns__maxPer)
                if soPer > self._localPeriodicPatterns__maxSoPer and tsPre - start >= self._localPeriodicPatterns__minDur:
                    PTL[item].add((start, tsPre))
                """else:
                    bitVector = 0b1 << currentTs+1
                    different = abs(self.tsList[item].bit_length() - bitVector.bit_length())
                    bitVector = bitVector | 0b1
                    bitVector = bitVector << different
                    self.tsList[item] = self.tsList[item] | bitVector"""
                if soPer <= self._localPeriodicPatterns__maxSoPer and self.__tsMax - start >= self._localPeriodicPatterns__minDur:
                    PTL[item].add((start, self.__tsMax))
                """else:
                    bitVector = 0b1 << currentTs+1
                    different = abs(self.tsList[item].bit_length() - bitVector.bit_length())
                    bitVector = bitVector | 0b1
                    bitVector = bitVector << different
                    self.tsList[item] = self.tsList[item] | bitVector"""
        self.__PTL = {k: v for k, v in PTL.items() if len(v) > 0}
        self.__items = list(self.__PTL.keys())

    def __createLPPTree(self) -> None:
        """
        Create transaction tree of local periodic item from input data.
        """
        for line in self.__Database:
            ts = int(line[0])
            tempTransaction = [item for item in line[1:] if item in self.__items]
            transaction = sorted(tempTransaction, key=lambda x: len(self.__PTL[x]), reverse=True)
            self.__root.addTransaction(transaction, ts)
            # for line in self.__Database:
            #     tid = int(transaction[0])
            #     tempTransaction = [item for item in transaction[1:] if item in self.__items]
            #     transaction = sorted(tempTransaction, key=lambda x: len(self.__PTL[x]), reverse=True)
            #     self.__root.addTransaction(transaction, tid)

    def __patternGrowth(self, tree: 'Tree', prefix: List[int], prefixPFList: Dict[Any, Any]) -> None:
        """
        Create prefix tree and prefixPFList. Store finalPatterns and its PTL.

        :param tree: The root node of prefix tree.
        :type tree: Node or Tree
        :param prefix: Prefix item list.
        :type prefix: list
        :param prefixPFList: tsList of prefix patterns.
        :type prefixPFList: dict or list
        :return: None
        """
        items = list(prefixPFList)
        if not prefix:
            items = reversed(items)
        for item in items:
            prefixCopy = prefix.copy()
            prefixCopy.append(item)
            PFList = {}
            prefixTree = Tree()
            prefixNode = tree.firstNodeLink[item]
            tidList = prefixNode.tidList
            path = []
            currentNode = prefixNode.parent
            while currentNode.item != -1:
                path.insert(0, currentNode.item)
                currentNodeItem = currentNode.item
                if currentNodeItem in PFList:
                    PFList[currentNodeItem] |= tidList
                else:
                    PFList[currentNodeItem] = tidList
                currentNode = currentNode.parent
            prefixTree.createPrefixTree(path, tidList)
            while prefixNode.nodeLink:
                prefixNode = prefixNode.nodeLink
                tidList = prefixNode.tidList
                path = []
                currentNode = prefixNode.parent
                while currentNode.item != -1:
                    path.insert(0, currentNode.item)
                    currentNodeItem = currentNode.item
                    if currentNodeItem in PFList:
                        PFList[currentNodeItem] = PFList[currentNodeItem] | tidList
                    else:
                        PFList[currentNodeItem] = tidList
                    currentNode = currentNode.parent
                prefixTree.createPrefixTree(path, tidList)
            if len(prefixCopy) == 1:
                self._localPeriodicPatterns__finalPatterns[prefixCopy[0]] = self.__calculatePTLbit(self.__tsList[item])
            else:
                self._localPeriodicPatterns__finalPatterns[tuple(prefixCopy)] = self.__calculatePTL(prefixPFList[item])
            candidateItems = list(PFList)
            for i in candidateItems:
                PTL = self.__calculatePTL(PFList[i])
                if len(PTL) == 0:
                    prefixTree.deleteNode(i)
                    del PFList[i]
            if PFList:
                self.__patternGrowth(prefixTree, prefixCopy, PFList)

    def __calculatePTL(self, tsList: List[int]) -> set:
        """
        Calculate PTL from input tsList as integer list

        :param tsList: It is tsList which store time stamp as integer.
        :type tsList: list
        :return: PTL
        :rtype: set
        """
        start = -1
        PTL = set()
        tsList = sorted(tsList)
        tsPre = tsList[0]
        soPer = ' '
        for ts in tsList[1:]:
            per = ts - tsPre
            if per <= self._localPeriodicPatterns__maxPer and start == -1:
                start = tsPre
                soPer = self._localPeriodicPatterns__maxSoPer
            if start != -1:
                soPer = max(0, soPer + per - self._localPeriodicPatterns__maxPer)
                if soPer > self._localPeriodicPatterns__maxSoPer:
                    if tsPre - start >= self._localPeriodicPatterns__minDur:
                        PTL.add((start, tsPre))
                    start = -1
            tsPre = ts
        if start != -1:
            soPer = max(0, soPer + self.__tsMax - tsPre - self._localPeriodicPatterns__maxPer)
            if soPer > self._localPeriodicPatterns__maxSoPer and tsPre - start >= self._localPeriodicPatterns__minDur:
                PTL.add((start, tsPre))
            if soPer <= self._localPeriodicPatterns__maxSoPer and self.__tsMax - start >= self._localPeriodicPatterns__minDur:
                PTL.add((start, self.__tsMax))
        return PTL

    def __calculatePTLbit(self, tsList: List[int]) -> set:
        """
        Calculate PTL from input tsList as bit vector.

        :param tsList: It is tsList which store time stamp as bit vector.
        :type tsList: list
        :return: PTL
        :rtype: set
        """
        tsList = list(bin(tsList))
        tsList = tsList[2:]
        start = -1
        currentTs = 1
        PTL = set()
        tsPre = ' '
        soPer = ' '
        for ts in tsList[currentTs:]:
            if ts == '0':
                currentTs += 1
                continue
            else:
                tsPre = currentTs
                currentTs += 1
                break
        for ts in tsList[currentTs:]:
            if ts == '0':
                currentTs += 1
                continue
            else:
                per = currentTs - tsPre
                if per <= self._localPeriodicPatterns__maxPer and start == -1:
                    start = tsPre
                    soPer = self._localPeriodicPatterns__maxSoPer
                if start != -1:
                    soPer = max(0, soPer + per - self._localPeriodicPatterns__maxPer)
                    if soPer > self._localPeriodicPatterns__maxSoPer:
                        if tsPre - start >= self._localPeriodicPatterns__minDur:
                            PTL.add((start, tsPre))
                        start = -1
                tsPre = currentTs
                currentTs += 1
        if start != -1:
            soPer = max(0, soPer + self.__tsMax - tsPre - self._localPeriodicPatterns__maxPer)
            if soPer > self._localPeriodicPatterns__maxSoPer and tsPre - start >= self._localPeriodicPatterns__minDur:
                PTL.add((start, tsPre))
            if soPer <= self._localPeriodicPatterns__maxSoPer and self.__tsMax - start >= self._localPeriodicPatterns__minDur:
                PTL.add((start, tsPre))
        return PTL

    def __convert(self, value: Any) -> float:
        """
        to convert the type of user specified minSup value

        :param value: user specified minSup value
        :type value: int or float or str
        :return: converted type
        :rtype: float
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self.__Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self.__Database) * value)
            else:
                value = int(value)
        return value

    @deprecated("It is recommended to use 'mine()' instead of 'mine()' for mining process. Starting from January 2025, 'mine()' will be completely terminated.")
    def startMine(self) -> None:
        """
        Mining process start from here.
        """
        self.mine()

    def mine(self) -> None:
        """
        Mining process start from here.
        """
        self._localPeriodicPatterns__startTime = _ab._time.time()
        self._localPeriodicPatterns__finalPatterns = {}
        self.__creatingItemSets()
        self._localPeriodicPatterns__maxPer = self.__convert(self._localPeriodicPatterns__maxPer)
        self._localPeriodicPatterns__maxSoPer = self.__convert(self._localPeriodicPatterns__maxSoPer)
        self._localPeriodicPatterns__minDur = self.__convert(self._localPeriodicPatterns__minDur)
        self.__createTSList()
        self.__generateLPP()
        self.__createLPPTree()
        self.__patternGrowth(self.__root, [], self.__items)
        self._localPeriodicPatterns__endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._localPeriodicPatterns__memoryUSS = float()
        self._localPeriodicPatterns__memoryRSS = float()
        self._localPeriodicPatterns__memoryUSS = process.memory_full_info().uss
        self._localPeriodicPatterns__memoryRSS = process.memory_info().rss

    def getMemoryUSS(self) -> float:
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._localPeriodicPatterns__memoryUSS

    def getMemoryRSS(self) -> float:
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._localPeriodicPatterns__memoryRSS

    def getRuntime(self) -> float:
        """
        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._localPeriodicPatterns__endTime - self._localPeriodicPatterns__startTime

    def getPatternsAsDataFrame(self) -> '_ab._pd.DataFrame':
        """
        Storing final local periodic patterns in a dataframe

        :return: returning local periodic patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._localPeriodicPatterns__finalPatterns.items():
            pat = str()
            for i in a:
                pat = pat + i + ' '
            data.append([pat, b])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'PTL'])
        return dataFrame

    def save(self, outFile: str) -> None:
        """
        Complete set of local periodic patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: csv file
        :return: None
        """
        self._localPeriodicPatterns__oFile = outFile
        writer = open(self._localPeriodicPatterns__oFile, 'w+')
        for x, y in self._localPeriodicPatterns__finalPatterns.items():
            pat = str()
            for i in x:
                pat = pat + i + '\t'
            pat = pat + ":"
            for i in y:
                pat = pat + str(i) + '\t'
            patternsAndPTL = pat.strip()
            writer.write("%s \n" % patternsAndPTL)

    def getPatterns(self) -> Dict:
        """
        Function to send the set of local periodic patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self._localPeriodicPatterns__finalPatterns

    def printResults(self) -> None:
        """
        This function is used to print the results
        """
        print("Total number of Local Periodic Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:",  self.getRuntime())


if __name__ == '__main__':
    _ap = str()
    if len(_ab._sys.argv) == 5 or len(_ab._sys.argv) == 6:
        if len(_ab._sys.argv) == 6:
            _ap = LPPGrowth(_ab._sys.argv[1], _ab._sys.argv[3], float(_ab._sys.argv[4]), _ab._sys.argv[5])
        if len(_ab._sys.argv) == 5:
            _ap = LPPGrowth(_ab._sys.argv[1], _ab._sys.argv[3], float(_ab._sys.argv[4]))
        _ap.mine()
        _ap.mine()
        print("Total number of Local Periodic Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in seconds:", _ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")



#  Copyright (C)  2021 Rage Uday Kiran
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC as _ABC, abstractmethod as _abstractmethod
import time as _time
import math as _math
import csv as _csv
import pandas as _pd
from collections import defaultdict as _defauldict
from itertools import combinations as _c
import os as _os
import os.path as _ospath
import psutil as _psutil
import sys as _sys
import validators as _validators
from urllib.request import urlopen as _urlopen


class _periodicFrequentPatterns(_ABC):
    """
    :Description:   This abstract base class defines the variables and methods that every periodic-frequent pattern mining algorithm must
                    employ in PAMI
    :Attributes:

        iFile : str
            Input file name or path of the input file
        maxPer: int or float or str
            The user can specify maxPer either in count or proportion of database size.
            If the program detects the data type of maxPer is integer, then it treats maxPer is expressed in count.
            Otherwise, it will be treated as float.
            Example: maxPer=10 will be treated as integer, while maxPer=10.0 will be treated as float
        k: int
            The user specify k in int
        sep : str
            This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
            However, the users can override their default separator.
        startTime:float
            To record the start time of the algorithm
        endTime:float
            To record the completion time of the algorithm
        finalPatterns: dict
            Storing the complete set of patterns in a dictionary variable
        oFile : str
            Name of the output file to store complete set of periodic-frequent patterns
        memoryUSS : float
            To store the total amount of USS memory consumed by the program
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program

    :Methods:

        mine()
            Mining process will start from here
        getPatterns()
            Complete set of patterns will be retrieved with this function
        storePatternsInFile(oFile)
            Complete set of periodic-frequent patterns will be loaded in to a output file
        getPatternsInDataFrame()
            Complete set of periodic-frequent patterns will be loaded in to data frame
        getMemoryUSS()
            Total amount of USS memory consumed by the program will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the program will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the program will be retrieved from this function
    """

    def __init__(self, iFile, k, maxPer, sep = '\t'):
        """
        :param iFile: Input file name or path of the input file
        :type iFile: str
        :param k: int or float
        :param maxPer: The user can specify maxPer either in count or proportion of database size.
            If the program detects the data type of maxPer is integer, then it treats maxPer is expressed in count.
            Otherwise, it will be treated as float.
            Example: maxPer=10 will be treated as integer, while maxPer=10.0 will be treated as float
        :type maxPer: int or float or str
        :param sep: separator used in user specified input file
        :type sep: str
        """

        self._iFile = iFile
        self._k = k
        self._maxPer = maxPer
        self._sep = sep
        self._oFile = str()
        self._startTime = float()
        self._endTime = float()
        self._memoryRSS = float()
        self._memoryUSS = float()
        self._finalPatterns = {}

    '''@abstractmethod
    def iFile(self):
        """Variable to store the input file path/file name"""

        pass

    @abstractmethod
    def k(self):
        """Variable to store the user-specified minimum support value"""

        pass

    @abstractmethod
    def maxPer(self):
        """Variable to store the user specified maximum periodicity value"""

        pass

    def sep(self):
        """Variable to store the separator in input file"""

        pass

    @abstractmethod
    def startTime(self):
        """Variable to store the start time of the mining process"""

        pass

    @abstractmethod
    def endTime(self):
        """Variable to store the end time of the complete program"""

        pass

    @abstractmethod
    def memoryUSS(self):
        """Variable to store the end time of the complete program"""

        pass

    @abstractmethod
    def memoryRSS(self):
        """Variable to store the end time of the complete program"""

        pass

    @abstractmethod
    def finalPatterns(self):
        """Variable to store the complete set of patterns in a dictionary"""

        pass

    @abstractmethod
    def oFile(self):
        """Variable to store the name of the output file to store the complete set of periodic-frequent patterns"""

        pass'''

    @_abstractmethod
    def startMine(self):
        """Code for the mining process will start from this function"""

        pass

    @_abstractmethod
    def getPatterns(self):
        """Complete set of periodic-frequent patterns generated will be retrieved from this function"""

        pass

    @_abstractmethod
    def getPatternsAsDataFrame(self):
        """Complete set of periodic-frequent patterns will be loaded in to data frame from this function"""

        pass

    @_abstractmethod
    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the program will be retrieved from this function"""

        pass

    @_abstractmethod
    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the program will be retrieved from this function"""
        pass

    @_abstractmethod
    def getRuntime(self):
        """Total amount of runtime taken by the program will be retrieved from this function"""

        pass

    @_abstractmethod
    def printResults(self):
        """ To print the Results of execution."""

        pass



# This code is used to calculate multiple minimum support of items in the the given database. Output can be stored in file or as as dataframe.
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#             from PAMI.extras.calculateMISValues import usingBeta as db
#
#             obj = db.usingBeta(iFile, 16, "\t")
#
#             obj.getPatternsAsDataFrame("outputFileName") # To create patterns as dataframes
#
#             obj.save(oFile)
#





__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import sys as _sys
import pandas as _pd
import validators as _validators
from urllib.request import urlopen as _urlopen

class usingBeta():
    """

    :Description: This code is used to calculate multiple minimum support of items in the the given database. Output can be stored in file or as as dataframe.
        
    :param  iFile: str :
                   Name of the Input file to get the patterns as DataFrame
    :param  beta: str :
                   Name of the output file to store complete set of frequent patterns
    :param  LS: int :
                   The user can specify threshold either in count or proportion of database size. If the program detects the data type of threshold is integer, then it treats threshold is expressed in count.
    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.

    **Importing this algorithm into a python program**
    --------------------------------------------------------
    .. code-block:: python

    from PAMI.extras.calculateMISValues import usingBeta as db

    obj = db.usingBeta(iFile, 3, 16, "\t")

    obj.save(oFile)

    """

    _iFile: str = ' '
    _beta: int = int()
    _sep: str = str()
    _LS: int = int()
    _finalPatterns: dict = {}

    def __init__(self, iFile: str, beta: int, LS: int, sep: str="\t"):
        self._iFile = iFile
        self._beta = beta
        self._LS = LS
        self._sep = sep
        self._lno = 0

    def _creatingItemSets(self) -> None:
        """
        Storing the complete transactions of the database/input file in a database variable
        """
        self._Database = []
        self._mapSupport = {}
        if isinstance(self._iFile, _pd.DataFrame):
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                self._Database = self._iFile['Transactions'].tolist()

        if isinstance(self._iFile, str):
            if _validators.url(self._iFile):
                data = _urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r') as f:
                        for line in f:
                            self._lno += 1
                            splitter = [i.rstrip() for i in line.split(self._sep)]
                            splitter = [x for x in splitter if x]
                            self._Database.append(splitter)
                except IOError:
                    print("File Not Found")

    def _creatingFrequentItems(self) -> tuple:
        """
        This function creates frequent items from _database.
        :return: frequentTidData that stores frequent items and their tid list.
        """
        tidData = {}
        self._lno = 0
        for transaction in self._Database:
            self._lno = self._lno + 1
            for item in transaction:
                    if item not in tidData:
                        tidData[item] = [self._lno]
                    else:
                        tidData[item].append(self._lno)
        #mini = min([len(k) for k in tidData.values()])
        frequentTidData = {k: len(v) * self._beta for k, v in tidData.items()}
        return frequentTidData

    def calculateMIS(self) -> None:
            self._creatingItemSets()
            frequentItems = self._creatingFrequentItems()
            for x, y in frequentItems.items():
              #self._finalPatterns[x] = min([y, self._LS])
                if y < self._LS:
                    self._finalPatterns[x] = self._LS
                else:
                    self._finalPatterns[x] = y

    def getMISDataFrame(self) -> _pd.DataFrame:
        """
        Storing items and its respective minimum support in a dataframe
        :return: returning items and its respective minimum support in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a, b])
            dataFrame = _pd.DataFrame(data, columns=['Items', 'MIS'])
        return dataFrame

    def save(self, outFile: str) -> None:
        """
        Complete set of items and its respective minimum support values will be loaded in to an output file
        :param outFile: name of the output file
        :type outFile: csv file
        :return: None
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            patternsAndSupport = x + "\t" + str(int(y))
            writer.write("%s \n" % patternsAndSupport)



if __name__ == '__main__':
  cd = usingBeta(sys.argv[1], sys.argv[3], sys.argv[4], sys.argv[5])
  cd.calculateMIS()
  cd.save(sys.argv[2])



import random as _rd
import sys as _sys


class createSyntheticUtility:
    """

    This class create synthetic utility database.

    :Attribute:

        totalTransactions : int
            No of transactions
        noOfItems : int 
            No of items
        maxUtilRange: int
            Maximum utility range
        avgTransactionLength : int
            The length of average transaction
        outputFile: str
            Name of the output file.

    :Methods:

        createUtilityDatabase(outputFile)
            Create utility database from DataFrame and store into outputFile

    **Credits:**
    ---------------
             The complete program was written by  P.Likhitha under the supervision of Professor Rage Uday Kiran.

    """
    
    def __init__(self, transactions: int, items: int, maxUtilRange: int, avgTransaction: int) -> None:
        self._totalTransactions = transactions
        self._noOfItems = items
        self._maxUtilRange = maxUtilRange
        self._avgTransactionLength = avgTransaction
    
    def createUtilityDatabase(self, outputFile: str) -> None:
        """
        create transactional database and return outputFileName

        :param outputFile: file name or path to store database
        :type outputFile: str
        :return: outputFile name
        """
        writer = open(outputFile, 'w+')
        for i in range(self._totalTransactions):
            length = _rd.randint(1, self._avgTransactionLength + 20)
            st = str()
            st1 = str()
            su = []
            for i in range(length):
                item = _rd.randint(1, self._noOfItems)
                utility = _rd.randint(1, self._maxUtilRange) 
                st = st + str(item) + '\t'
                su.append(utility)
                st1 = st1 + str(utility) + '\t'
            summation = sum([i for i in su])
            st = st + ":" + str(summation) + ":"
            writer.write("%s" % st)
            writer.write("%s \n" % st1)
            
if __name__ == "__main__":
    _ap = str()
    _ap = createSyntheticUtility(100000, 870, 100, 10)
    _ap.createUtilityDatabase("T10_util.txt")
else:
    print("Error! The number of input parameters do not match the total number of parameters provided")



# gSpan is a subgraph mining algorithm that uses DFS and DFS codes to mine subgraphs
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#             from PAMI.subgraphMining.basic import gspan as alg
#
#             obj = alg.GSpan(iFile, minSupport)
#
#             obj.mine()
#
#             frequentGraphs = obj.getFrequentSubgraphs()
#
#             memUSS = obj.getMemoryUSS()
#
#             obj.save(oFile)
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#



from PAMI.subgraphMining.basic import abstract as _ab

class GSpan(_ab._gSpan):

    eliminate_infrequent_vertices = True
    eliminate_infrequent_vertex_pairs = True
    eliminate_infrequent_edge_labels = True
    edge_count_pruning = True

    def __init__(self, iFile, minSupport, outputSingleVertices=True, maxNumberOfEdges=float('inf'), outputGraphIds=False) -> None:
        """
        Initialize variables
        """
        
        self.minSup = minSupport
        self.frequentSubgraphs = []
        self._runtime = 0
        self.maxMem = 0
        self.graphCount = 0
        self.patternCount = 0
        self.frequentVertexLabels = []
        self.infrequentVerticesRemovedCount = 0
        self.infrequentVertexPairsRemoved = 0
        self.edgeRemovedByLabel = 0
        self.eliminatedWithMaxSize = 0
        self.emptyGraphsRemoved = 0
        self.pruneByEdgeCount = 0
        self.inPath = iFile
        self.outPath = None
        self.outputSingleVertices = outputSingleVertices
        self.maxNumberOfEdges = maxNumberOfEdges
        self.outputGraphIds = outputGraphIds
        self._memoryUSS = float()
        self._memoryRSS = float()

        self.label_mapping = {}
        self.current_label = 0
        self.edge_label_mapping = {}
        self.current_edge_label = 0


    def mine(self):

        if self.maxNumberOfEdges <= 0:
            return
        
        self.frequentSubgraphs = []

        self.patternCount = 0

        # Record the start time
        t1 = _ab.time.time()

        # Read graphs
        graphDb = self.readGraphs(self.inPath)

        # Calculate minimum support as a number of graphs
        self.minSup = _ab.math.ceil(self.minSup * len(graphDb))

        # Mining
        self.gSpan(graphDb, self.outputSingleVertices)

        # Output
        # self.writeResultToFile(self.outPath)

        t2 = _ab.time.time()

        #Calculate runtime
        self._runtime = (t2 - t1)

        process = _ab._psutil.Process(_ab._os.getpid())

        self._memoryUSS = float()

        self._memoryRSS = float()

        self._memoryUSS = process.memory_full_info().uss

        self._memoryRSS = process.memory_info().rss

        self.patternCount = len(self.frequentSubgraphs)


    def save(self, oFile):
        """
        The `save` function writes information about frequent subgraphs to a specified
        output file in a specific format.
        
        :param oFile: The `save` method is used to write the results of frequent
        subgraphs to a file specified by the `outputPath` parameter. The method iterates over each
        frequent subgraph in `self.frequentSubgraphs` and writes the subgraph information to the file
        """
        reverse_label_mapping = {v: k for k, v in self.label_mapping.items()}
        reverse_edge_label_mapping = {v: k for k, v in self.edge_label_mapping.items()}

        with open(oFile, 'w') as bw:
            i = 0
            for subgraph in self.frequentSubgraphs:
                sb = []

                dfsCode = subgraph.dfsCode
                sb.append(f"t # {i} * {subgraph.support}\n")
                if dfsCode.size == 1:
                    ee = dfsCode.getEeList()[0]
                    vLabel1 = reverse_label_mapping.get(ee.vLabel1, ee.vLabel1)
                    sb.append(f"v 0 {vLabel1}\n")
                    if ee.edgeLabel != -1:
                        vLabel2 = reverse_label_mapping.get(ee.vLabel2, ee.vLabel2)
                        edgeLabel = reverse_edge_label_mapping.get(ee.edgeLabel, ee.edgeLabel)
                        sb.append(f"v 1 {vLabel2}\n")
                        sb.append(f"e 0 1 {edgeLabel}\n")
                else:
                    vLabels = [reverse_label_mapping.get(label, label) for label in dfsCode.getAllVLabels()]
                    for j, vLabel in enumerate(vLabels):
                        sb.append(f"v {j} {vLabel}\n")
                    for ee in dfsCode.getEeList():
                        edgeLabel = reverse_edge_label_mapping.get(ee.edgeLabel, ee.edgeLabel)
                        sb.append(f"e {ee.v1} {ee.v2} {edgeLabel}\n")

                if self.outputGraphIds:
                    sb.append("x " + " ".join(str(id) for id in subgraph.setOfGraphsIds))

                sb.append("\n\n")
                bw.write("".join(sb))
                i += 1

    def readGraphs(self, path):
        """
        The `readGraphs` function reads graph data from a file and constructs a list of graphs with vertices
        and edges.
        
        :param path: The `path` parameter in the `readGraphs` method is the file path to the text file
        containing the graph data that needs to be read and processed.
        :return: The `readGraphs` method reads graph data from a file specified by the `path` parameter. It
        parses the data to create a list of graph objects and returns this list. Each graph object contains
        information about vertices and edges within the graph.
        """
        with open(path, 'r') as br:
            graphDatabase = []
            vMap = {}
            gId = None

            for line in br:
                line = line.strip()

                if line.startswith("t"):
                    if vMap:  # If vMap is not empty, it means a graph was read
                        graphDatabase.append(_ab.Graph(gId, vMap))
                        vMap = {}  # Reset for the next graph

                    gId = int(line.split(" ")[2])

                elif line.startswith("v"):
                    items = line.split(" ")
                    vId = int(items[1])
                    # Map vertex label
                    if items[2].isdigit():
                        vLabel = int(items[2])
                    else:
                        vLabel = self.get_label(items[2])
                    vMap[vId] = _ab.Vertex(vId, vLabel)

                elif line.startswith("e"):
                    items = line.split(" ")
                    v1 = int(items[1])
                    v2 = int(items[2])
                    # Map edge label
                    if items[3].isdigit():
                        eLabel = int(items[3])
                    else:
                        eLabel = self.get_edge_label(items[3])
                    e = _ab.Edge(v1, v2, eLabel)
                    vMap[v1].addEdge(e)
                    vMap[v2].addEdge(e)

            if vMap:
                graphDatabase.append(_ab.Graph(gId, vMap))

        self.graphCount = len(graphDatabase)
        return graphDatabase

    def get_label(self, label_char):
        if label_char not in self.label_mapping:
            self.label_mapping[label_char] = self.current_label
            self.current_label += 1
        return self.label_mapping[label_char]

    def get_edge_label(self, label_char):
        if label_char not in self.edge_label_mapping:
            self.edge_label_mapping[label_char] = self.current_edge_label
            self.current_edge_label += 1
        return self.edge_label_mapping[label_char]

    def subgraphIsomorphisms(self, c: _ab.DFSCode, g: _ab.Graph):
        """
        The function `subgraphIsomorphisms` takes a DFS code and a graph as input, and finds all subgraph
        isomorphisms between the DFS code and the graph.
        
        :param c: The parameter `c` in the `subgraphIsomorphisms` function is of type `_ab.DFSCode`, which
        seems to represent a Depth-First Search code.
        :param g: The parameter `g` in the `subgraphIsomorphisms` function represents a graph object. The
        function is trying to find subgraph isomorphisms between a given DFS code `c` and the graph `g`. It
        iterates through the vertices of the graph starting with a specific
        :return: The function `subgraphIsomorphisms` returns a list of dictionaries, where each dictionary
        represents a subgraph isomorphism mapping between the input DFS code `c` and the input graph `g`.
        Each dictionary in the list maps vertex IDs from the DFS code to corresponding vertex IDs in the
        graph, indicating a valid subgraph isomorphism.
        """
        isoms = []    
        startLabel = c.getEeList()[0].getVLabel1()

        # Find all vertices in the graph that match the start label and initialize isomorphisms with them
        for vId in g.findAllWithLabel(startLabel):
            hMap = {}
            hMap[0] = vId
            isoms.append(hMap)

        # For each edge in the DFS code, try to extend each partial isomorphism
        for ee in c.getEeList():
            v1, v2, v2Label, eLabel = ee.getV1(), ee.getV2(), ee.getVLabel2(), ee.getEdgeLabel() 
            updateIsoms = []
            # Try to extend each current isomorphism with the current edge
            for iso in isoms:
                mappedV1 = iso.get(v1)
                # Forward edge
                if v1 < v2:
                    mappedVertices = list(iso.values())
                    for mappedV2 in g.getAllNeighbors(mappedV1):
                        if (v2Label == mappedV2.getLabel() and
                            mappedV2.getId() not in mappedVertices and
                                eLabel == g.getEdgeLabel(mappedV1, mappedV2.getId())):

                            tempM = iso.copy()
                            tempM[v2] = mappedV2.getId()

                            updateIsoms.append(tempM)

                # Backward edge
                else:
                    mappedV2 = iso.get(v2)
                    # Check if the backward edge exists in the graph matching the DFS code edge
                    if g.isNeighboring(mappedV1, mappedV2) and eLabel == g.getEdgeLabel(mappedV1, mappedV2):
                        updateIsoms.append(iso)

            isoms = updateIsoms
        return isoms

    
    def rightMostPathExtensionsFromSingle(self, c: _ab.DFSCode, g: _ab.Graph):
        """
        The function `rightMostPathExtensionsFromSingle` generates extensions for a given DFS code and
        graph, focusing on the rightmost path.
        
        :param c: The parameter `c` is of type `_ab.DFSCode`, which seems to represent a Depth-First Search
        code. It is used in the `rightMostPathExtensionsFromSingle` method to perform operations related to
        DFS codes
        :param g: The parameter `g` in the provided code snippet represents a graph object. It seems to be
        an instance of a graph data structure that contains vertices and edges. The code is designed to
        find and return extensions from a given DFS code `c` based on the provided graph `g`. The function
        `
        :return: The function `rightMostPathExtensionsFromSingle` returns a dictionary `extensions`
        containing extended edges as keys and sets of graph IDs as values.
        """
        # Get the unique identifier for the given graph
        gid = g.getId()
        # Initialize a dictionary to store potential extensions
        extensions = {}

        # If the DFS code is empty, consider all edges of the graph for extension
        if c.isEmpty():
            for vertex in g.vertices:
                for e in vertex.getEdgeList():
                    # Determine the order of vertex labels to maintain consistency
                    v1Label = g.getVLabel(e.v1)
                    v2Label = g.getVLabel(e.v2)
                    if v1Label < v2Label:
                        ee1 = _ab.ExtendedEdge(0, 1, v1Label, v2Label, e.getEdgeLabel())
                    else:
                        ee1 = _ab.ExtendedEdge(0, 1, v2Label, v1Label, e.getEdgeLabel())

                    # Update the extensions dictionary with new or existing extended edges
                    setOfGraphIds = extensions.get(ee1, set())
                    setOfGraphIds.add(gid)
                    extensions[ee1] = setOfGraphIds
        else:
            # For non-empty DFS code, focus on extending from the rightmost path
            rightMost = c.getRightMost()
            isoms = self.subgraphIsomorphisms(c, g)

            # Iterate through all isomorphisms to find valid extensions
            for isom in isoms:
                invertedIsom = {v: k for k, v in isom.items()}
                mappedRm = isom[rightMost]
                mappedRmLabel = g.getVLabel(mappedRm)
                for x in g.getAllNeighbors(mappedRm):
                    invertedX = invertedIsom.get(x.getId())
                    if invertedX is not None and c.onRightMostPath(invertedX) and c.notPreOfRm(invertedX) and not c.containEdge(rightMost, invertedX):
                        ee = _ab.ExtendedEdge(rightMost, invertedX, mappedRmLabel, x.getLabel(), g.getEdgeLabel(mappedRm, x.getId()))
                        extensions.setdefault(ee, set()).add(gid)

                mappedVertices = set(isom.values())
                for v in c.getRightMostPath():
                    mappedV = isom[v]
                    mappedVLabel = g.getVLabel(mappedV)
                    for x in g.getAllNeighbors(mappedV):
                        if x.getId() not in mappedVertices:
                            ee = _ab.ExtendedEdge(v, rightMost + 1, mappedVLabel, x.getLabel(), g.getEdgeLabel(mappedV, x.getId()))
                            extensions.setdefault(ee, set()).add(gid)

        return extensions


    def rightMostPathExtensions(self, c: _ab.DFSCode, graphDb, graphIds):
        """
        The function `rightMostPathExtensions` generates extensions for a given DFS code by considering
        rightmost paths in a graph database.
        
        :param c: The parameter `c` in the `rightMostPathExtensions` method is of type `_ab.DFSCode`. It
        seems to represent a Depth-First Search code used in graph algorithms. The method is responsible
        for generating extensions based on the rightmost path in a graph
        :param graphDb: The `graphDb` parameter in the `rightMostPathExtensions` method is a
        database that stores graph data. It is used to retrieve graph objects based on
        their IDs, which are provided in the `graphIds` parameter. The method then performs operations on
        these graph objects to generate
        :param graphIds: The `graphIds` parameter in the `rightMostPathExtensions` function represents a
        list of graph identifiers. These identifiers are used to retrieve specific graphs from the
        `graphDb` database in order to perform operations on them within the function. Each ID in the
        `graphIds` list corresponds to an identifier.
        :return: The function `rightMostPathExtensions` returns a dictionary `extensions` containing
        extended edges as keys and sets of graph IDs as values.
        """
        extensions = {}
        if c.isEmpty():
            for id in graphIds:
                g = graphDb[id]
                # Skip graphs if pruning based on edge count is enabled and applicable
                if GSpan.edge_count_pruning and c.size >= g.getEdgeCount():
                    self.pruneByEdgeCount += 1
                    continue
                for v in g.vertices:
                    for e in v.getEdgeList():
                        # Organize the vertex labels to maintain consistent ordering
                        v1L = g.getVLabel(e.v1)
                        v2L = g.getVLabel(e.v2)
                        if v1L < v2L:
                            ee1 = _ab.ExtendedEdge(0, 1, v1L, v2L, e.getEdgeLabel())
                        else:
                            ee1 = _ab.ExtendedEdge(0, 1, v2L, v1L, e.getEdgeLabel())

                        # Add the new or existing extensions to the dictionary                       
                        setOfGraphIds = extensions.get(ee1, set())
                        setOfGraphIds.add(id)
                        extensions[ee1] = setOfGraphIds
        else:
            # For non-empty DFS codes, extend based on the rightmost path of each graph
            rightMost = c.getRightMost()
            for id in graphIds:
                g = graphDb[id]
                if GSpan.edge_count_pruning and c.size >= g.getEdgeCount():
                    self.pruneByEdgeCount += 1
                    continue
                isoms = self.subgraphIsomorphisms(c, g)
                for isom in isoms:
                    invertedIsom = {}
                    for key, value in isom.items():
                        invertedIsom[value] = key
                    mappedRM = isom.get(rightMost)
                    mappedRMLabel = g.getVLabel(mappedRM)
                    for x in g.getAllNeighbors(mappedRM):
                        invertedX = invertedIsom.get(x.getId())
                        if invertedX is not None and c.onRightMostPath(invertedX) and \
                        c.notPreOfRm(invertedX) and not c.containEdge(rightMost, invertedX):
                            
                            ee = _ab.ExtendedEdge(rightMost, invertedX, mappedRMLabel, x.getLabel(),
                                            g.getEdgeLabel(mappedRM, x.getId()))

                            if ee not in extensions:
                                extensions[ee] = set()
                            extensions[ee].add(g.getId())

                    mappedVertices = isom.values()
                    for v in c.getRightMostPath():
                        mappedV = isom[v]
                        mappedVLabel = g.getVLabel(mappedV)
                        for x in g.getAllNeighbors(mappedV):
                            if x.getId() not in mappedVertices:
                                ee = _ab.ExtendedEdge(v, rightMost + 1, mappedVLabel, x.getLabel(),
                                                g.getEdgeLabel(mappedV, x.getId()))

                                if ee not in extensions:
                                    extensions[ee] = set()
                                extensions[ee].add(g.getId())
        return extensions



    def gspanDFS(self, c: _ab.DFSCode, graphDb, subgraphId):
        """
        The `gspanDFS` function recursively explores graph patterns using the gSpan algorithm to find
        frequent subgraphs in a graph database.
        
        :param c: In the provided code snippet, the parameter `c` is an instance of the `_ab.DFSCode` class.
        It is used as an input to the `gspanDFS` method for performing Depth-First Search (DFS) traversal in
        a graph mining algorithm. The `c` parameter represents
        :type c: _ab.DFSCode
        :param graphDb: The `graphDb` parameter  refers to a graph database that the algorithm is 
        operating on.
        :param subgraphId: The `subgraphId` parameter in the `gspanDFS` method refers to an
        ID represents a specific subgraph within the graph database `graphDb`. 
        :return: The `gspanDFS` method is a recursive function that is called within itself to explore the graph 
        structure and find frequent subgraphs. The function does not have a return value, but it modifies 
        the `self.frequentSubgraphs` list by appending new frequent subgraphs found during the DFS traversal.
        """

        if c.size == self.maxNumberOfEdges - 1:
            return
        extensions = self.rightMostPathExtensions(c, graphDb, subgraphId)

        for extension, newGraphIds in extensions.items():
            sup = len(newGraphIds)
            
            if (sup >= self.minSup):
                newC = c.copy()
                newC.add(extension)
                
                if (self.isCanonical(newC)):
                    subgraph = _ab.FrequentSubgraph(newC, newGraphIds, sup)
                    self.frequentSubgraphs.append(subgraph)

                    self.gspanDFS(newC, graphDb, newGraphIds)


    def isCanonical(self, c: _ab.DFSCode):
        """
        The function `isCanonical` checks if a given DFS code is canonical by comparing it with its
        rightmost path extensions.
        
        :param c: The parameter `c` is an instance of the `_ab.DFSCode` class
        :type c: _ab.DFSCode
        :return: a boolean value. It returns True if the input DFSCode `c` is canonical, and False if it is
        not canonical.
        """
        canC = _ab.DFSCode()
        for i in range(c.size):
            extensions = self.rightMostPathExtensionsFromSingle(canC, _ab.Graph(-1, None, c))
            minEe = None
            for ee in extensions.keys():
                if minEe is None or ee.smallerThan(minEe):
                    minEe = ee

            if minEe is not None and minEe.smallerThan(c.getAt(i)):
                return False
            
            if minEe is not None:
                canC.add(minEe)
        return True
    

    def gSpan(self, graphDb, outputSingleVertices):
        """
        The gSpan function in Python processes a graph database by precalculating vertex lists, removing
        infrequent vertex pairs, and performing a depth-first search algorithm.
        
        :param graphDb: The `graphDb` parameter  refers to a graph database that the algorithm is 
        operating on.
        :param outputSingleVertices: The `outputFrequentVertices` parameter is a boolean flag that
        determines whether single vertices should be output or not.
        """
        if outputSingleVertices or GSpan.eliminate_infrequent_vertices:
            self.findAllOnlyOneVertex(graphDb, outputSingleVertices)

        for g in graphDb:
            g.precalculateVertexList()

        if GSpan.eliminate_infrequent_vertex_pairs or GSpan.eliminate_infrequent_edge_labels:
            self.removeInfrequentVertexPairs(graphDb)

        graphIds = set()
        for i, g in enumerate(graphDb):
            if g.vertices is not None and len(g.vertices) != 0:
                if self.infrequentVerticesRemovedCount > 0:
                    g.precalculateVertexList()

                graphIds.add(i)
                g.precalculateVertexNeighbors()
                g.precalculateLabelsToVertices()
            else:
                self.emptyGraphsRemoved += 1

        if len(self.frequentVertexLabels) != 0:
            self.gspanDFS(_ab.DFSCode(), graphDb, graphIds)


    class Pair:
        def __init__(self, x, y):
            if x < y:
                self.x = x
                self.y = y
            else:
                self.x = y
                self.y = x

        def __eq__(self, other):
            if isinstance(other, GSpan.Pair):
                return self.x == other.x and self.y == other.y
            return False

        def __hash__(self):
            return self.x + 100 * self.y


    def findAllOnlyOneVertex(self, graphDb, outputFrequentVertices):
        """
        The function `findAllOnlyOneVertex` iterates through a graph database to find frequent vertices
        based on a minimum support threshold, storing the results and optionally removing infrequent
        vertices.
        
        :param graphDb: The `graphDb` parameter  refers to a graph database that the algorithm is 
        operating on.
        :param outputFrequentVertices: The `outputFrequentVertices` parameter is a boolean flag that
        determines whether single vertices should be included in the output or not.
        """
        self.frequentVertexLabels = []
        labelM = {} 
        for g in graphDb:
            for v in g.getNonPrecalculatedAllVertices():
                if v.getEdgeList():
                    vLabel = v.getLabel()
                    labelM.setdefault(vLabel, set()).add(g.getId())
        # Check each label for frequency against the minimum support threshold
        for label, tempSupG in labelM.items():                
            sup = len(tempSupG)
            if sup >= self.minSup:
                self.frequentVertexLabels.append(label)
                if outputFrequentVertices:
                    tempD = _ab.DFSCode()
                    tempD.add(_ab.ExtendedEdge(0, 0, label, label, -1))
                    self.frequentSubgraphs.append(_ab.FrequentSubgraph(tempD, tempSupG, sup))
            elif GSpan.eliminate_infrequent_vertices:
                for graphId in tempSupG:
                    g = graphDb[graphId]
                    g.removeInfrequentLabel(label)
                    self.infrequentVerticesRemovedCount += 1


    def removeInfrequentVertexPairs(self, graphDb):
        """
        The function `removeInfrequentVertexPairs` processes a graph database by removing infrequent vertex
        pairs and edge labels based on specified support thresholds.
        
        :param graphDb: The `graphDb` parameter  refers to a graph database that the algorithm is 
        operating on.
        """
        if GSpan.eliminate_infrequent_edge_labels:
            matrix = _ab.SparseTriangularMatrix()
            alreadySeenPair = set() # To avoid double counting pairs in the same graph

        if GSpan.eliminate_infrequent_edge_labels:
            mapEdgeLabelToSupport = {}
            alreadySeenEdgeLabel = set() # To avoid double counting edge labels in the same graph

        for g in graphDb:
            vertices = g.getAllVertices()

            # Check each vertex and its edges for infrequent pairs and labels
            for v1 in vertices:
                labelV1 = v1.getLabel()

                for edge in v1.getEdgeList():
                    v2 = edge.another(v1.getId())
                    labelV2 = g.getVLabel(v2)

                    # Track vertex label pairs for infrequency analysis
                    if GSpan.eliminate_infrequent_edge_labels:
                        pair = self.Pair(labelV1, labelV2)
                        if pair not in alreadySeenPair:
                            matrix.incrementCount(labelV1, labelV2)
                            alreadySeenPair.add(pair)

                    # Track edge labels for infrequency analysis
                    if GSpan.eliminate_infrequent_edge_labels:
                        edgeLabel = edge.getEdgeLabel()
                        if edgeLabel not in alreadySeenEdgeLabel:
                            alreadySeenEdgeLabel.add(edgeLabel)
                            edgeSupport = mapEdgeLabelToSupport.get(edgeLabel, 0)
                            mapEdgeLabelToSupport[edgeLabel] = edgeSupport + 1

            if GSpan.eliminate_infrequent_vertex_pairs:
                alreadySeenPair.clear()
            if GSpan.eliminate_infrequent_edge_labels:
                alreadySeenEdgeLabel.clear()

        if GSpan.eliminate_infrequent_vertex_pairs:
            matrix.removeInfrequentEntriesFromMatrix(self.minSup)

        if GSpan.eliminate_infrequent_vertex_pairs or GSpan.eliminate_infrequent_edge_labels:
            for g in graphDb:
                vertices = g.getAllVertices()

                for v1 in vertices:
                    iterEdges = iter(v1.getEdgeList())
                    for edge in iterEdges:
                        v2 = edge.another(v1.getId())
                        labelV2 = g.getVLabel(v2)
                        count = matrix.getSupportForItems(v1.getLabel(), labelV2)

                        # Remove edges based on infrequency criteria
                        if GSpan.eliminate_infrequent_vertex_pairs and count < self.minSup:
                            v1.removeEdge(edge)
                            self.infrequentVertexPairsRemoved += 1

                        elif GSpan.eliminate_infrequent_edge_labels and \
                                mapEdgeLabelToSupport.get(edge.getEdgeLabel(), 0) < self.minSup:
                            v1.removeEdge(edge)
                            self.edgeRemovedByLabel += 1
    

    def getMemoryRSS(self):
        return self._memoryRSS

    def getMemoryUSS(self):
        return self._memoryUSS

    def getRuntime(self):
        return self._runtime
    
    def getFrequentSubgraphs(self):
        sb = []  
        for i, subgraph in enumerate(self.frequentSubgraphs):  
            dfsCode = subgraph.dfsCode
            subgraphDescription = [f"t # {i} * {subgraph.support}"]  
            
            if dfsCode.size == 1:
                ee = dfsCode.getEeList()[0]
                subgraphDescription.append(f"v 0 {ee.vLabel1}")
                if ee.edgeLabel != -1:
                    subgraphDescription.append(f"v 1 {ee.vLabel2}")
                    subgraphDescription.append(f"e 0 1 {ee.edgeLabel}")
            else:
                vLabels = dfsCode.getAllVLabels()
                for j, vLabel in enumerate(vLabels):
                    subgraphDescription.append(f"v {j} {vLabel}")
                for ee in dfsCode.getEeList():
                    subgraphDescription.append(f"e {ee.v1} {ee.v2} {ee.edgeLabel}")
            
            sb.append('\n'.join(subgraphDescription))  
        return '\n'.join(sb)  

    def getSubgraphGraphMapping(self):
        """
        Return a list of mappings from subgraphs to the graph IDs they belong to in the format <FID, Clabel, GIDs[]>.
        """
        mappings = []
        for i, subgraph in enumerate(self.frequentSubgraphs):
            mapping = {
                "FID": i,
                "Clabel": str(subgraph.dfsCode),
                "GIDs": list(subgraph.setOfGraphsIds)
            }
            mappings.append(mapping)
        return mappings

    def saveSubgraphsByGraphId(self, oFile):
        """
        Save subgraphs by graph ID as a flat transaction, such that each row represents the graph ID and each row can contain multiple subgraph IDs.
        """
        graphToSubgraphs = {}
        
        for i, subgraph in enumerate(self.frequentSubgraphs):
            for graphId in subgraph.setOfGraphsIds:
                if graphId not in graphToSubgraphs:
                    graphToSubgraphs[graphId] = []
                graphToSubgraphs[graphId].append(i)

        graphToSubgraphs = {k: graphToSubgraphs[k] for k in sorted(graphToSubgraphs)}

        with open(oFile, 'w') as f:
            for _, subgraphIds in graphToSubgraphs.items():
                f.write(f"{' '.join(map(str, subgraphIds))}\n")




# PFPGrowth is one of the fundamental algorithm to discover periodic-frequent patterns in a transactional database.
#
# **Importing this algorithm into a python program**
#
#             from PAMI.periodicFrequentPattern.basic import PFPGrowth as alg
#
#             iFile = 'sampleDB.txt'
#
#             minSup = 10  # can also be specified between 0 and 1
#
#             maxPer = 20 # can also be specified between 0 and 1
#
#             obj = alg.PFPGrowth(iFile, minSup, maxPer)
#
#             obj.mine()
#
#             periodicFrequentPatterns = obj.getPatterns()
#
#             print("Total number of Periodic Frequent Patterns:", len(periodicFrequentPatterns))
#
#             obj.save("periodicFrequentPatterns")
#
#             Df = obj.getPatternsAsDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#


__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""

from PAMI.periodicFrequentPattern.basic import abstract as _ab
from typing import Dict, Tuple
import pandas as pd
from deprecated import deprecated
import numpy as np

_maxPer = float()
_minSup = float()
_lno = int()


class _Node(object):
    """
    A class used to represent the node of frequentPatternTree

    :**Attributes**:    - **item** (*int or None*) -- *Storing item of a node.*
                        - **timeStamps** (*list*) -- *To maintain the timestamps of a database at the end of the branch.*
                        - **parent** (*list*) -- *To maintain the parent of every node.*
                        - **children** (*list*) -- *To maintain the children of a node.*

    :**Methods**:    -**addChild(itemName)** -- *Storing the children to their respective parent nodes.*
    """

    def __init__(self, item, locations, parent=None):
        self.item = item
        self.locations = locations
        self.parent = parent
        self.children = {}

    def addChild(self, item, locations):
        """
        This method takes an item and locations as input, adds a new child node
        if the item does not already exist among the current node's children, or
        updates the locations of the existing child node if the item is already present.

        :param item: Represents the distinct item to be added as a child node.
        :type item: Any
        :param locations: Represents the locations associated with the item.
        :type locations: list
        :return: The child node associated with the item.
        :rtype: _Node
        """
        if item not in self.children:
            self.children[item] = _Node(item, locations, self)
        else:
            self.children[item].locations = locations + self.children[item].locations
            
        return self.children[item]

    def traverse(self):
        """
        This method constructs a transaction by traversing from the current node to the root node, collecting items along the way.

        :return: A tuple containing the transaction and the locations associated with the current node.
        :rtype: tuple(list, Any)
        """
        transaction = []
        locs = self.locations
        node = self.parent
        while node.parent is not None:
            transaction.append(node.item)
            node = node.parent
        return transaction[::-1], locs


class PFPGrowth(_ab._periodicFrequentPatterns):
    """
    **About this algorithm**

    :**Description**:   PFPGrowth is one of the fundamental algorithm to discover periodic-frequent patterns in a transactional database.

    :**Reference**:   Syed Khairuzzaman Tanbeer, Chowdhury Farhan, Byeong-Soo Jeong, and Young-Koo Lee, "Discovering Periodic-Frequent
                      Patterns in Transactional Databases", PAKDD 2009, https://doi.org/10.1007/978-3-642-01307-2_24


    :**Parameters**:    - **iFile** (*str or URL or dataFrame*) -- *Name of the Input file to mine complete set of frequent patterns.*
                        - **oFile** (*str*) -- *Name of the output file to store complete set of frequent patterns.*
                        - **minSup** (*int or float or str*) -- *The user can specify minSup either in count or proportion of database size. If the program detects the data type of minSup is integer, then it treats minSup is expressed in count. Otherwise, it will be treated as float.*
                        - **maxPer** (*int or float or str*) -- *The user can specify maxPer either in count or proportion of database size. It controls the maximum number of transactions in which any two items within a pattern can reappear.*
                        - **sep** (*str*) -- *This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.*

    :**Attributes**:    - **startTime** (*float*) -- *To record the start time of the mining process.*
                        - **endTime** (*float*) -- *To record the completion time of the mining process.*
                        - **finalPatterns** (*dict*) -- *Storing the complete set of patterns in a dictionary variable.*
                        - **memoryUSS** (*float*) -- *To store the total amount of USS memory consumed by the program.*
                        - **memoryRSS** (*float*) -- *To store the total amount of RSS memory consumed by the program.*
                        - **Database** (*list*) -- *To store the transactions of a database in list.*
                        - **mapSupport** (*Dictionary*) -- *To maintain the information of item and their frequency.*
                        - **lno** (*int*) -- *It represents the total no of transactions*
                        - **tree** (*class*) -- *it represents the Tree class.*
                        - **itemSetCount** (*int*) -- *it represents the total no of patterns.*

    :**Methods**:       - **mine()** -- *Mining process will start from here.*
                        - **getPatterns()** -- *Complete set of patterns will be retrieved with this function.*
                        - **save(oFile)** -- *Complete set of periodic-frequent patterns will be loaded in to a output file.*
                        - **getPatternsAsDataFrame()** -- *Complete set of periodic-frequent patterns will be loaded in to a dataframe.*
                        - **getMemoryUSS()** -- *Total amount of USS memory consumed by the mining process will be retrieved from this function.*
                        - **getMemoryRSS()** -- *Total amount of RSS memory consumed by the mining process will be retrieved from this function.*
                        - **getRuntime()** -- *Total amount of runtime taken by the mining process will be retrieved from this function.*
                        - **creatingItemSets(fileName)** -- *Scans the dataset and stores in a list format.*
                        - **PeriodicFrequentOneItem()** -- *Extracts the one-periodic-frequent patterns from database.*
                        - **updateDatabases()** -- *Update the database by removing aperiodic items and sort the Database by item decreased support.*
                        - **buildTree()** -- *After updating the Database, remaining items will be added into the tree by setting root node as null.*
                        - **convert()** -- *This methos is used to convert the user specified value.*

    **Execution methods**

    **Terminal command**

    .. code-block:: console

       Format:

       (.venv) $ python3 PFPGrowth.py <inputFile> <outputFile> <minSup> <maxPer>

       Example usage:

       (.venv) $ python3 PFPGrowth.py sampleDB.txt patterns.txt 10.0 20.0

    .. note:: minSup will be considered in percentage of database transactions


    **Calling from a python program**

    .. code-block:: python

            from PAMI.periodicFrequentPattern.basic import PFPGrowth as alg

            iFile = 'sampleDB.txt'

            minSup = 10  # can also be specified between 0 and 1

            maxPer = 20 # can also be specified between 0 and 1

            obj = alg.PFPGrowth(iFile, minSup, maxPer)

            obj.mine()

            periodicFrequentPatterns = obj.getPatterns()

            print("Total number of Periodic Frequent Patterns:", len(periodicFrequentPatterns))

            obj.save("periodicFrequentPatterns")

            Df = obj.getPatternsAsDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)


    **Credits**

    The complete program was written by P. Likhitha  and revised by Tarun Sreepada under the supervision of Professor Rage Uday Kiran.

    """
    _startTime = float()
    _endTime = float()
    _minSup = str()
    _maxPer = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []
    _rank = {}
    _rankedUp = {}
    _lno = 0

    def _creatingItemSets(self) -> None:
        """
        Storing the complete transactions of the database/input file in a database variable

        :return: None
        """
        self._Database = []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            data, ts = [], []
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'TS' in i:
                ts = self._iFile['TS'].tolist()
            if 'Transactions' in i:
                data = self._iFile['Transactions'].tolist()
            for i in range(len(data)):
                if data[i]:
                    tr = [str(ts[i])] + [x for x in data[i].split(self._sep)]
                    self._Database.append(tr)
                else:
                    self._Database.append([str(ts[i])])

        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self._Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()

    def _convert(self, value) -> int:
        """
        To convert the given user specified value

        :param value: user specified value
        :return: converted value
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self._Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self._Database) * value)
            else:
                value = int(value)
        return value

    @deprecated("It is recommended to use 'mine()' instead of 'mine()' for mining process. Starting from January 2025, 'mine()' will be completely terminated.")
    def startMine(self) -> None:
        self.mine()

    def _getMaxPer(self, arr, maxTS):
        """
        This method appends `0` and `maxTS` to the input array, sorts the array,
        and then computes the differences between consecutive elements.

        :param arr: The input array of elements.
        :type arr: numpy.ndarray
        :param maxTS: The maximum timestamp to be appended to the array.
        :type maxTS: int or float
        :return: None
        """
        arr = np.append(arr, [0, maxTS])
        arr = np.sort(arr)
        arr = np.diff(arr)

        return np.max(arr)

    def _construct(self, items, data, minSup, maxPer, maxTS, patterns):

        """
        This method filters the items based on the minimum support (minSup) and
        maximum period (maxPer). It then constructs a tree structure from the
        filtered items and data.

        :param items: A dictionary where keys are items and values are lists of timestamps.
        :type items: dict
        :param data: The dataset used to construct the tree, where each entry is a list with
                     an index followed by items.
        :type data: list of lists
        :param minSup: The minimum support threshold.
        :type minSup: int
        :param maxPer: The maximum period threshold.
        :type maxPer: int or float
        :param maxTS: The maximum timestamp.
        :type maxTS: int or float
        :param patterns: A dictionary to store the patterns discovered during the construction.
        :type patterns: dict
        :return: A tuple containing the root node of the constructed tree and a dictionary
                 of item nodes.
        :rtype: tuple(_Node, dict)
        """

        # maxPerItems = {k: self.getMaxPer(v, maxTS) for k, v in items.items() if len(v) >= minSup}

        items = {k: v for k, v in items.items() if len(v) >= minSup and self._getMaxPer(v, maxTS) <= maxPer}

        #tested ok
        for item, ts in items.items():
            # pat = "\t".join(item)
            # self.patCount += 1
            # patterns[pat] = (len(ts), self.getMaxPer(ts, maxTS))
            patterns[tuple([item])] = [len(ts), self._getMaxPer(ts, maxTS)]

        root = _Node([], None, None)
        itemNodes = {}
        for line in data:
            currNode = root
            index = int(line[0])
            line = line[1:]
            line = sorted([item for item in line if item in items], key = lambda x: len(items[x]), reverse = True)
            for item in line:
                currNode = currNode.addChild(item, [index])   # heavy
                if item in itemNodes:
                    itemNodes[item].add(currNode)
                else:
                    itemNodes[item] = set([currNode])

        return root, itemNodes


    def _recursive(self, root, itemNode, minSup, maxPer, patterns, maxTS):
        """
        This method recursively constructs a pattern tree from the given root node,
        filtering items based on the minimum support (minSup) and maximum period (maxPer).
        It updates the patterns dictionary with the discovered patterns.

        :param root: The current root node of the pattern tree.
        :type root: _Node
        :param itemNode: A dictionary where keys are items and values are sets of nodes
                         associated with those items.
        :type itemNode: dict
        :param minSup: The minimum support threshold.
        :type minSup: int
        :param maxPer: The maximum period threshold.
        :type maxPer: int or float
        :param patterns: A dictionary to store the patterns discovered during the recursion.
        :type patterns: dict
        :param maxTS: The maximum timestamp.
        :type maxTS: int or float
        """

        for item in itemNode:
            newRoot = _Node(root.item + [item], None, None)

            itemLocs = {}
            transactions = {}
            for node in itemNode[item]:
                transaction, locs = node.traverse()
                if len(transaction) < 1:
                    continue
                # transactions.append((transaction, locs))
                if tuple(transaction) in transactions:
                    transactions[tuple(transaction)].extend(locs)
                else:
                    transactions[tuple(transaction)] = locs

                for item in transaction:
                    if item in itemLocs:
                        itemLocs[item] += locs
                    else:
                        itemLocs[item] = list(locs)

            # Precompute getMaxPer results for itemLocs
            maxPerResults = {item: self._getMaxPer(itemLocs[item], maxTS) for item in itemLocs if len(itemLocs[item]) >= minSup}

            # Filter itemLocs based on minSup and maxPer
            itemLocs = {k: len(v) for k, v in itemLocs.items() if k in maxPerResults and maxPerResults[k] <= maxPer}

            # Iterate over filtered itemLocs
            for item in itemLocs:
                # pat = "\t".join([str(x) for x in newRoot.item + [item]])
                # self.patCount += 1
                # patterns[pat] = [itemLocs[item], maxPerResults[item]]
                patterns[tuple(newRoot.item + [item])] = [itemLocs[item], maxPerResults[item]]
            
            if not itemLocs:
                continue

            newItemNodes = {}

            for transaction, locs in transactions.items():
                transaction = sorted([item for item in transaction if item in itemLocs], key = lambda x: itemLocs[x], reverse = True)
                if len(transaction) < 1:
                    continue
                currNode = newRoot
                for item in transaction:
                    currNode = currNode.addChild(item, locs)
                    if item in newItemNodes:
                        newItemNodes[item].add(currNode)
                    else:
                        newItemNodes[item] = set([currNode])

            self._recursive(newRoot, newItemNodes, minSup, maxPer, patterns, _lno)

    def mine(self) -> None:
        """
        Mining process will start from this function

        :return: None
        """

        global _minSup, _maxPer, _lno
        self._startTime = _ab._time.time()
        if self._iFile is None:
            raise Exception("Please enter the file path or file name:")
        if self._minSup is None:
            raise Exception("Please enter the Minimum Support")
        if self._maxPer is None:
            raise Exception("Please enter the Maximum Periodicity")
        if self._sep is None:
            raise Exception("Default separator is tab space, please enter the separator if you have different separator in the input file")

        self._creatingItemSets()
        self._minSup = self._convert(self._minSup)
        self._maxPer = self._convert(self._maxPer)
        #tested ok
        _minSup, _maxPer, _lno = self._minSup, self._maxPer, len(self._Database)
        if self._minSup > len(self._Database):
            raise Exception("Please enter the minSup in range between 0 to 1")
        

        items = {}

        # tested ok
        for line in self._Database:
            index = int(line[0])
            for item in line[1:]:
                if item not in items:
                    items[item] = []
                items[item].append(index)

        root, itemNodes = self._construct(items, self._Database, _minSup, _maxPer, _lno, self._finalPatterns)

        self._recursive(root, itemNodes, _minSup, _maxPer, self._finalPatterns, _lno)

        newPattern = {}
        for k, v in self._finalPatterns.items():
            newPattern["\t".join([str(x) for x in k])] = v

        self._finalPatterns = newPattern
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Periodic Frequent patterns were generated successfully using PFPGrowth algorithm ")

    def getMemoryUSS(self) -> float:
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self) -> float:
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self) -> float:
        """
        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self) -> _ab._pd.DataFrame:
        """
        Storing final periodic-frequent patterns in a dataframe

        :return: returning periodic-frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a, b[0], b[1]])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support', 'Periodicity'])
        return dataFrame

    def save(self, outFile: str) -> None:
        """
        Complete set of periodic-frequent patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: csv file
        :return: None
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = x + ":" + str(y[0]) + ":" + str(y[1])
            #s1 = x.replace(' ', '\t').strip() + ":" + str(y[0]) + ":" + str(y[1])
            writer.write("%s \n" % s1)

    def getPatterns(self) -> Dict[str, Tuple[int, int]]:
        """
        Function to send the set of periodic-frequent patterns after completion of the mining process

        :return: returning periodic-frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self) -> None:
        """
        This function is used to print the results

        :return: None
        """
        print("Total number of Periodic Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:",  self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 5 or len(_ab._sys.argv) == 6:
        if len(_ab._sys.argv) == 6:
            _ap = PFPGrowth(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5])
        if len(_ab._sys.argv) == 5:
            _ap = PFPGrowth(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        _ap.mine()
        _ap.mine()
        print("Total number of Periodic-Frequent Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")



import pandas as pd
import numpy as np
import sys
import time
import os
import psutil

class TemporalDatabase:

    def __init__(self, databaseSize: int,
                 avgItemsPerTransaction: int,
                 numItems: int,
                 sep: str = '\t',
                 occurrenceProbabilityOfSameTimestamp: float = 0.1,
                 occurrenceProbabilityToSkipSubsequentTimestamp: float = 0.1) -> None:

        self.databaseSize = databaseSize
        self.avgItemsPerTransaction = avgItemsPerTransaction
        self.numItems = numItems
        self.sep = sep
        self.occurrenceProbabilityOfSameTimestamp = occurrenceProbabilityOfSameTimestamp
        self.occurrenceProbabilityToSkipSubsequentTimestamp = occurrenceProbabilityToSkipSubsequentTimestamp

    def performCoinFlip(self, probability: float) -> bool:
        """
        Perform a coin flip with the given probability.

        :param probability: Probability of the coin landing heads (i.e., the event occurring).
        :return: True if the coin lands heads, False otherwise.
        """
        result = np.random.choice([0, 1], p=[1 - probability, probability])
        return result == 1

    def tuning(self, array, sumRes) -> list:
        """
        Tune the array to ensure that the sum of the values equals sumRes.
        """
        num_transactions = len(array)
        # Initialize values randomly between 1 and numItems
        values = np.random.randint(1, self.numItems + 1, num_transactions)

        # normalize values to ensure sum equals sumRes
        values = values / np.sum(values) * sumRes
        values = np.round(values).astype(int)


        # Adjust values to ensure sum equals sumRes
        while np.sum(values) != sumRes:
            current_sum = np.sum(values)
            if current_sum > sumRes:
                # Decrease the value of a random index
                indices = np.where(values > 1)[0]
                if len(indices) == 0:
                    raise ValueError("Cannot adjust values to meet sumRes")
                idx = np.random.choice(indices)
                values[idx] -= 1
            else:
                # Increase the value of a random index
                indices = np.where(values < self.numItems)[0]
                if len(indices) == 0:
                    raise ValueError("Cannot adjust values to meet sumRes")
                idx = np.random.choice(indices)
                values[idx] += 1

        # Assign adjusted values back to array
        for i in range(len(array)):
            array[i][1] = values[i]

        return array

    def create(self) -> None:
        """
        Create the temporal database or DataFrame based on the specified type of file.
        """
        start = time.time()

        self.db = []
        lineSize = []


        self.current_timestamp = 0  # Initialize current timestamp
        sumRes = self.databaseSize * self.avgItemsPerTransaction  # Total number of items

        for i in range(self.databaseSize):
            # Determine the timestamp
            if self.performCoinFlip(self.occurrenceProbabilityOfSameTimestamp):
                timestamp = self.current_timestamp
            else:
                if self.performCoinFlip(self.occurrenceProbabilityToSkipSubsequentTimestamp):
                    self.current_timestamp += 2
                else:
                    self.current_timestamp += 1
                timestamp = self.current_timestamp

            self.db.append([timestamp])  # Start the transaction with the timestamp

            lineSize.append([i, 0])  # Initialize lineSize with 0 for each transaction

        # Adjust lineSize to ensure sum of sizes equals sumRes
        lineSize = self.tuning(lineSize, sumRes)

        # For each transaction, generate items
        for i in range(len(lineSize)):
            transaction_index = lineSize[i][0]
            num_items = lineSize[i][1]

            if num_items > self.numItems:
                raise ValueError(
                    "Error: Either increase numItems or decrease avgItemsPerTransaction or modify percentage")
            items = np.random.choice(range(1, self.numItems + 1), num_items, replace=False)
            self.db[transaction_index].extend(items)

        self._runTime = time.time() - start
        process = psutil.Process(os.getpid())
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss

    def save(self, outputFile: str = None) -> None:
        """
        Save the temporal database to the specified output file.
        """
        if outputFile is not None:
            self.outputFile = outputFile
        else:
            self.outputFile = "temporalDatabase.txt"

        with open(self.outputFile, 'w') as writer:
            for line in self.db:
                writer.write(self.sep.join(map(str, line)) + '\n')

    def getRuntime(self) -> float:
        """
        Returns the runtime of the algorithm in seconds.
        """
        return self._runTime
    
    def getMemoryRSS(self) -> int:
        """
        """
        return self._memoryRSS
    
    def getMemoryUSS(self) -> int:
        """
        """
        return self._memoryUSS

    def getTransactions(self) -> None:
        """
        Convert the database to a DataFrame.
        """
        # merge all the transactions into a single DataFrame
        timestamps = []
        transactions = []

        for line in self.db:
            timestamps.append(line[0])
            transactions.append(line[1:])

        self.df = pd.DataFrame([timestamps, transactions], index=['Timestamp', 'Items']).T

        return self.df

if __name__ == '__main__':
    if len(sys.argv) == 7:
        obj = TemporalDatabase(
            databaseSize=int(sys.argv[1]),
            avgItemsPerTransaction=int(sys.argv[2]),
            numItems=int(sys.argv[3]),
            outputFile=str(sys.argv[4]),
            occurrenceProbabilityOfSameTimestamp=float(sys.argv[5]),
            occurrenceProbabilityToSkipSubsequentTimestamp=float(sys.argv[6]),
            sep=sys.argv[7]
        )
        obj.create()
        obj.save()
    else:
        print("Usage: python TemporalDatabase.py <databaseSize> <avgItemsPerTransaction> <numItems> <outputFile> <occurrenceProbabilityOfSameTimestamp> <occurrenceProbabilityToSkipSubsequentTimestamp> <sep> ")

        obj = TemporalDatabase(
            databaseSize=100000,
            avgItemsPerTransaction=10,
            numItems=50,
            sep="\t",
            occurrenceProbabilityOfSameTimestamp=0.1,
            occurrenceProbabilityToSkipSubsequentTimestamp=0.1
        )
        obj.create()
        obj.save("temporalDatabase.txt")

        print(obj.getTransactions())

        sys.exit(1)


# ECLATDiffest uses diffset to extract the frequent patterns in a transactional database.
#
# **Importing this algorithm into a python program**
#
#             import PAMI.frequentPattern.basic.ECLATDiffset as alg
#
#             iFile = 'sampleDB.txt'
#
#             minSup = 10  # can also be specified between 0 and 1
#
#             obj = alg.ECLATDiffset(iFile, minSup)
#
#             obj.mine()
#
#             frequentPatterns = obj.getPatterns()
#
#             print("Total number of Frequent Patterns:", len(frequentPatterns))
#
#             obj.savePatterns(oFile)
#
#             Df = obj.getPatternInDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#


__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


# from abstract import *

from PAMI.frequentPattern.basic import abstract as _ab
from deprecated import deprecated


class ECLATDiffset(_ab._frequentPatterns):
    """
    :**Description**:   ECLATDiffset uses diffset to extract the frequent patterns in a transactional database.

    :**Reference**:  KDD '03: Proceedings of the ninth ACM SIGKDD international conference on Knowledge discovery and data mining
                     August 2003 Pages 326–335 https://doi.org/10.1145/956750.956788
            
    :**Parameters**:    - **iFile** (*str or URL or dataFrame*) -- *Name of the Input file to mine complete set of frequent patterns.*
                        - **oFile** (*str*) -- *Name of the output file to store complete set of frequent patterns*
                        - **minSup** (*int or float or str*) -- *The user can specify minSup either in count or proportion of database size. If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.*
                        - **sep** (*str*) -- **This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.**

    :**Attributes**:    - **startTime** (*float*) -- *To record the start time of the mining process.*
                        - **endTime** (*float*) -- *To record the end time of the mining process.*
                        - **finalPatterns** (*dict*) -- *Storing the complete set of patterns in a dictionary variable.*
                        - **memoryUSS** (*float*) -- *To store the total amount of USS memory consumed by the program.*
                        - **memoryRSS** *(float*) -- *To store the total amount of RSS memory consumed by the program.*
                        - **Database** (*list*) -- *To store the transactions of a database in list.*
          
        
    Execution methods
    =================

    **Terminal command**

    .. code-block:: console

      Format:

      (.venv) $ python3 ECLATDiffset.py <inputFile> <outputFile> <minSup>

      Example Usage:

      (.venv) $ python3 ECLATDiffset.py sampleDB.txt patterns.txt 10.0

    .. note:: minSup can be specified  in support count or a value between 0 and 1.

    
     **Calling from a python program**

    .. code-block:: python

            import PAMI.frequentPattern.basic.ECLATDiffset as alg

            iFile = 'sampleDB.txt'

            minSup = 10  # can also be specified between 0 and 1

            obj = alg.ECLATDiffset(iFile, minSup)

            obj.mine()

            frequentPatterns = obj.getPatterns()

            print("Total number of Frequent Patterns:", len(frequentPatterns))

            obj.savePatterns(oFile)

            Df = obj.getPatternInDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)


    Credits:
    ========

    The complete program was written by Kundai and revised by Tarun Sreepada under the supervision of Professor Rage Uday Kiran.

    """

    _minSup = float()
    _startTime = float()
    _endTime = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []
    _diffSets = {}
    _trans_set = set()

    def _creatingItemSets(self):
        """
        Storing the complete transactions of the database/input file in a database variable
        """
        self._Database = []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                self._Database = self._iFile['Transactions'].tolist()
        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self._Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()

    def _convert(self, value):
        """

        To convert the user specified minSup value

        :param value: user specified minSup value
        :return: converted type
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self._Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self._Database) * value)
            else:
                value = int(value)
        return value

    def _getUniqueItemList(self):

        # tidSets will store all the initial tids
        tidSets = {}
        # uniqueItem will store all frequent 1 items
        uniqueItem = []
        for line in self._Database:
                transNum = 0
                # Database = [set([i.rstrip() for i in transaction.split('\t')]) for transaction in f]
                for transaction in self._Database:
                    transNum += 1
                    self._trans_set.add(transNum)
                    for item in transaction:
                        if item in tidSets:
                            tidSets[item].add(transNum)
                        else:
                            tidSets[item] = {transNum}
        for key, value in tidSets.items():
            supp = len(value)
            if supp >= self._minSup:
                self._diffSets[key] = [supp, self._trans_set.difference(value)]
                uniqueItem.append(key)
        # for x, y in self._diffSets.items():
        #     print(x, y)
        uniqueItem.sort()
        # print()
        return uniqueItem

    def _runDeclat(self, candidateList):
        """

        It will generate the combinations of frequent items

        :param candidateList :it represents the items with their respective transaction identifiers
        :type candidateList: list
        :return: returning transaction dictionary
        :rtype: dict
        """

        newList = []
        for i in range(0, len(candidateList)):
            item1 = candidateList[i]
            iList = item1.split()
            for j in range(i + 1, len(candidateList)):
                item2 = candidateList[j]
                jList = item2.split()
                if iList[:-1] == jList[:-1]:
                    unionDiffSet = self._diffSets[item2][1].difference(self._diffSets[item1][1])
                    unionSup = self._diffSets[item1][0] - len(unionDiffSet)
                    if unionSup >= self._minSup:
                        newKey = item1 + "\t" + jList[-1]
                        self._diffSets[newKey] = [unionSup, unionDiffSet]
                        newList.append(newKey)
                    else: 
                        break

        if len(newList) > 0:
            self._runDeclat(newList)

    @deprecated("It is recommended to use 'mine()' instead of 'mine()' for mining process. Starting from January 2025, 'mine()' will be completely terminated.")
    def startMine(self):
        """
        Frequent pattern mining process will start from here
        """
        self.mine()

    def mine(self):
        """
        Frequent pattern mining process will start from here
        """

        self._startTime = _ab._time.time()
        self._Database = []
        self._finalPatterns = {}
        self._diffSets = {}
        self._trans_set = set()
        if self._iFile is None:
            raise Exception("Please enter the file path or file name:")
        if self._minSup is None:
            raise Exception("Please enter the Minimum Support")
        self._creatingItemSets()
        #print(len(self._Database))
        self._minSup = self._convert(self._minSup)
        uniqueItemList = []
        uniqueItemList = self._getUniqueItemList()
        self._runDeclat(uniqueItemList)
        self._finalPatterns = self._diffSets
        #print(len(self._finalPatterns), len(uniqueItemList))
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Frequent patterns were generated successfully using ECLAT Diffset algorithm")

    def getMemoryUSS(self):
        """

        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """

        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """

        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """

        Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a.replace('\t', ' '), b[0]])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataFrame

    def save(self, outFile):
        """

        Complete set of frequent patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: csvfile
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            patternsAndSupport = x.strip() + ":" + str(y[0])
            writer.write("%s \n" % patternsAndSupport)

    def getPatterns(self):
        """

        Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self):
        """
        This function is used to print the results
        """
        print("Total number of Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:",  self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 4 or len(_ab._sys.argv) == 5:
        if len(_ab._sys.argv) == 5:
            _ap = ECLATDiffset(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        if len(_ab._sys.argv) == 4:
            _ap = ECLATDiffset(_ab._sys.argv[1], _ab._sys.argv[3])
        _ap.mine()
        _ap.mine()
        print("Total number of Frequent Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print(_ap.getPatternsAsDataFrame())
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")

